from __future__ import annotations

import enum
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import simpy

from simulatte.typings import ProcessGenerator

if TYPE_CHECKING:
    from simulatte.system import System


class ArmPosition(enum.Enum):
    AT_PICKUP = 0
    AT_RELEASE = 1


class Robot(simpy.Resource):
    def __init__(self, *, system: System, pick_timeout: float, place_timeout: float, rotation_timeout: float) -> None:
        super().__init__(env=system.env, capacity=1)

        self.env = system.env
        self.pick_timeout = pick_timeout
        self.place_timeout = place_timeout
        self.rotation_timeout = rotation_timeout
        self.arm_position = ArmPosition.AT_PICKUP
        self._worked_time = 0
        self._movements = 0
        self._saturation_history: list[tuple[float, float]] = [(0, 0)]
        self._productivity_history: list[tuple[float, float]] = [(0, 0)]

    @property
    def worked_time(self) -> int:
        return self._worked_time

    @property
    def productivity(self) -> float:
        return self._movements / self.env.now

    @worked_time.setter
    def worked_time(self, value: int) -> None:
        self._worked_time = value
        self._saturation_history.append((self.env.now, self.saturation))

    @property
    def saturation(self) -> float:
        return self._worked_time / self.env.now

    @property
    def idle_time(self) -> int:
        return self.env.now - self.worked_time

    def _pick_process(self) -> ProcessGenerator:
        if self.arm_position == ArmPosition.AT_RELEASE:
            yield self.rotate()

        yield self.env.timeout(self.pick_timeout)
        self.worked_time += self.pick_timeout
        self.arm_position = ArmPosition.AT_PICKUP

    def pick(self) -> simpy.Process:
        return self.env.process(self._pick_process())

    def _place_process(self) -> ProcessGenerator:
        if self.arm_position == ArmPosition.AT_PICKUP:
            yield self.rotate()

        yield self.env.timeout(self.place_timeout)
        self.worked_time += self.place_timeout
        self.arm_position = ArmPosition.AT_RELEASE
        self._movements += 1
        self._productivity_history.append((self.env.now, self.productivity))

    def place(self) -> simpy.Process:
        return self.env.process(self._place_process())

    def _rotate_process(self) -> ProcessGenerator:
        yield self.env.timeout(self.rotation_timeout)
        self.worked_time += self.rotation_timeout

    def rotate(self) -> simpy.Process:
        return self.env.process(self._rotate_process())

    def plot(self, *, show_productivity=False) -> None:
        x = [t / 60 / 60 for t, _ in self._saturation_history]
        y = [s * 100 for _, s in self._saturation_history]
        plt.plot(x, y)
        plt.xlabel("Time [h]")
        plt.ylabel(f"Saturation [%]")
        plt.title("Robot Productivity")
        plt.ylim([0, 100])
        plt.show()

        if show_productivity:
            x = [t for t, _ in self._productivity_history]
            y = [p * 60 * 60 for _, p in self._productivity_history]
            plt.plot(x, y)
            plt.title("Productivity [pcs/h]")
            plt.show()



from __future__ import annotations

from typing import TYPE_CHECKING

from ..areas import InternalArea
from .base_observer import Observer

if TYPE_CHECKING:
    from simulatte.operations import FeedingOperation
    from simulatte.picking_cell.areas.position import Position


class InternalObserver(Observer[InternalArea]):
    def next(self) -> FeedingOperation | None:
        return min(self.observable_area.cell.staging_area, default=None)

    def _can_enter(self, *, feeding_operation: FeedingOperation) -> tuple[bool, Position | None, Position | None]:
        for unload_position in self.cell.internal_area.unload_positions:
            if not unload_position.busy:
                return True, None, unload_position
        return False, None, None

    def _main_process(self) -> None:
        """
        Manage the shift of a FeedingOperation from the StagingArea to the InternalArea.

        Called when the InternalArea signal event is triggered (see StagingObserver._main_process).

        It checks if the staging area is not empty and the internal area is not full.

        If all conditions are met, the procedure follows the following steps:
        1. Removes the FeedingOperation from the StagingArea.
        2. Register the FeedingOperation into the InternalArea.
        3. Updates the status of the FeedingOperation.
        4. Initialize the process that will take care of the Ant logistic movements.
        """

        if (
            not self.cell.staging_area.is_empty
            and not self.cell.internal_area.is_full
            and (feeding_operation := self.next()) is not None
        ):

            can_enter, pre_unload_position, unload_position = self._can_enter(feeding_operation=feeding_operation)

            if can_enter:

                feeding_operation.pre_unload_position = pre_unload_position
                feeding_operation.unload_position = unload_position

                # Remove the FeedingOperation from the StagingArea
                self.cell.staging_area.remove(feeding_operation)

                # Register the FeedingOperation into the InternalArea
                self.cell.internal_area.append(feeding_operation)

                feeding_operation.ant.enter_internal_area()

                # Start moving the ant to the unloading position
                self.cell.let_ant_in(feeding_operation=feeding_operation)



from __future__ import annotations

import math
from typing import TYPE_CHECKING, Literal, cast, Type

import simulatte
from simulatte.ant import Ant
from simulatte.exceptions import OutOfStockError
from simulatte.products import Product, ProductsGenerator
from simulatte.requests import Request
from simulatte.stores import WarehouseStore
from simulatte.stores.warehouse_location import PhysicalPosition
from simulatte.system.policies import LocationPolicy, UnitLoadPolicy
from simulatte.unitload import Pallet, Tray

if TYPE_CHECKING:
    from simulatte.stores.warehouse_location.warehouse_location import WarehouseLocation
    from simulatte import System


class StoresManager:
    def __init__(self, *, unit_load_policy: UnitLoadPolicy, location_policy: LocationPolicy):
        self._stores: list[WarehouseStore] = []
        self._stores: dict[Type[WarehouseStore], list] = {}

        self._unit_load_policy = unit_load_policy
        self._location_policy = location_policy  # to be used to find locations for storing of products
        self._stock: dict[int, dict[str, dict[str, int]]] = {}
        self.system: System | None = None

        self._magic = {"pallet": [0, 0], "tray": [0, 0]}
        self._ip_history = []

    def __call__(self, store: WarehouseStore) -> None:
        """
        Register a store to be managed by the StoresManager.
        """
        self._stores.setdefault(type(store), []).append(store)

    def register_system(self, system: System) -> None:
        self.system = system

    @property
    def stores(self) -> dict[Type[WarehouseStore], list]:
        return self._stores

    @staticmethod
    def freeze(location: WarehouseLocation, unit_load: Pallet) -> None:
        location.freeze(unit_load=unit_load)

    def update_stock(
        self,
        *,
        product: Product,
        case_container: Literal["pallet", "tray"],
        inventory: Literal["on_hand", "on_transit"],
        n_cases: int,
    ) -> None:
        """
        Modify the stock (on hand and on transit) quantities of a product.
        """

        if product.id not in self._stock:
            self._stock[product.id] = {
                "pallet": {"on_hand": 0, "on_transit": 0},
                "tray": {"on_hand": 0, "on_transit": 0},
            }

        self._stock[product.id][case_container][inventory] += n_cases

        pallet_on_hand = 0
        pallet_on_transit = 0
        tray_on_hand = 0
        tray_on_transit = 0

        for product_id in [81]:
            try:
                pallet_on_hand += self._stock[product_id]["pallet"]["on_hand"]
                pallet_on_transit += self._stock[product_id]["pallet"]["on_transit"]
                tray_on_hand += self._stock[product_id]["tray"]["on_hand"]
                tray_on_transit += self._stock[product_id]["tray"]["on_transit"]

                self._ip_history.append(
                    {
                        "time": self.system.env.now,
                        "pallet_on_hand": pallet_on_hand,
                        "pallet_on_transit": pallet_on_transit,
                        "tray_on_hand": tray_on_hand,
                        "tray_on_transit": tray_on_transit,
                    }
                )
            except KeyError:
                pass

    def inventory_position(self, *, product: Product, case_container: Literal["pallet", "tray"]) -> int:
        """
        Return the inventory position of a product, filtered in pallets of trays.
        """
        on_hand = self._stock[product.id][case_container]["on_hand"]
        on_transit = self._stock[product.id][case_container]["on_transit"]
        return on_hand + on_transit

    @simulatte.as_process
    def load(self, *, store: WarehouseStore, ant: Ant) -> None:
        """
        Used to centralize the loading of unitloads into the stores.
        Needed to keep trace of the on hand quantity of each product,
        to trigger replenishment when needed.

        This method must be called when the ant is in front of the store, waiting to be
        unloaded by the store.

        It triggers the loading process of the store.
        """

        from eagle_trays.asrs import ASRS
        from eagle_trays.avsrs import AVSRS

        # troviamo la locazione per il pallet/vassoio
        location = self.get_location_for_unit_load(store=store, unit_load=ant.unit_load)

        if isinstance(store, ASRS):
            case_container = cast(Literal, "pallet")
        elif isinstance(store, AVSRS):
            case_container = cast(Literal, "tray")
        else:
            raise ValueError(f"Case container {type(ant.unit_load)} not supported.")

        # riduciamo l'on_transit
        self.update_stock(
            product=ant.unit_load.product,
            case_container=case_container,
            inventory="on_transit",
            n_cases=-ant.unit_load.n_cases,
        )

        # alziamo l'on_hand
        self.update_stock(
            product=ant.unit_load.product,
            case_container=case_container,
            inventory="on_hand",
            n_cases=ant.unit_load.n_cases,
        )

        yield store.load(unit_load=ant.unit_load, location=location, ant=ant, priority=10)

    def unload(
        self, *, type_of_stores: Type[WarehouseStore], picking_request: Request
    ) -> tuple[tuple[WarehouseStore, WarehouseLocation, PhysicalPosition],...]:
        """
        Used to centralize the unloading of unitloads from the stores.
        Needed to keep trace of the on hand quantity of each product.
        It does not trigger replenishment operations.

        This method should be called when the system is organizing the feeding operation.
        It does NOT trigger the unloading process of the store.
        """

        from eagle_trays.asrs import ASRS
        from eagle_trays.avsrs import AVSRS

        if type_of_stores is ASRS:
            case_container = cast(Literal, "pallet")
        elif type_of_stores is AVSRS:
            case_container = cast(Literal, "tray")
        else:
            raise ValueError

        # Get location
        stores = self.stores[type_of_stores]
        stores_and_locations = self.find_stores_locations_for_output(
            stores=stores,
            product=picking_request.product,
            quantity=picking_request.n_cases,
            raise_on_none=False,
        )

        for (store, location, position) in stores_and_locations:

            location.book_pickup(position.unit_load)

            # aumentiamo l'on_transit
            self.update_stock(
                product=picking_request.product,
                case_container=case_container,
                inventory="on_transit",
                n_cases=max(0, position.unit_load.n_cases - picking_request.n_cases),
            )

            # riduciamo l'on_hand
            self.update_stock(
                product=picking_request.product,
                case_container=case_container,
                inventory="on_hand",
                n_cases=-position.unit_load.n_cases,
            )

            # controlliamo necessità di replenishment
            self.check_replenishment(product=picking_request.product, case_container=case_container)

        return stores_and_locations

    def check_replenishment(
        self,
        *,
        product: Product,
        case_container: Literal["pallet", "tray"],
        periodic_check=False,
    ):
        """
        Checks if there is need for replenishment operations.
        Used both in the unload method and in the
        periodic replenishment process.
        """

        inventory_position = self.inventory_position(product=product, case_container=case_container)
        s_max = product.s_max[case_container]
        s_min = product.s_min[case_container]

        if periodic_check or inventory_position <= s_min:
            # calcoliamo quanti cases ci servono per arrivare a S_max
            n_cases = s_max - inventory_position
            n_cases = max(0, n_cases)
            n_pallet = math.ceil(n_cases / product.case_per_pallet)

            # aumentiamo l'on_transit
            self.update_stock(
                product=product,
                case_container=case_container,
                inventory="on_transit",
                n_cases=n_pallet * product.case_per_pallet,
            )

            for _ in range(n_pallet):
                self.system.store_replenishment(
                    product=product,
                    case_container=case_container,
                )

    @simulatte.as_process
    def periodic_store_replenishment(self):
        """
        Periodically checks if there is need for replenishment operations.
        """
        while True:
            yield self.system.env.timeout(60 * 60 * 8)  # TODO: mettere come parametro

            for product in self.system.products:
                for case_container in ("pallet", "tray"):
                    self.check_replenishment(
                        product=product,
                        case_container=case_container,
                        periodic_check=True,
                    )

    def find_location_for_product(self, *, store: WarehouseStore, product: Product) -> WarehouseLocation:
        return self._location_policy(store=store, product=product)

    def get_location_for_unit_load(self, *, store: WarehouseStore, unit_load: Pallet) -> WarehouseLocation:
        """
        FOR INPUT.

        Find a location for a unit load in a store.
        Find the location accordingly to the LocationPolicy set.
        Then freeze the location to prevent other unit loads from
        being placed in the same location.
        """
        location = self.find_location_for_product(store=store, product=unit_load.product)
        store.book_location(location=location, unit_load=unit_load)
        return location

    def find_stores_locations_for_output(
        self,
        *,
        stores: list[WarehouseStore],
        product: Product,
        quantity: int,
        raise_on_none: bool = False,
    ) -> tuple[tuple[WarehouseStore, WarehouseLocation, PhysicalPosition],...]:
        """
        FOR OUTPUT.

        Get a tuple of stores and locations from which to pickup a product.
        """
        try:
            stores_and_locations = self._unit_load_policy(stores=stores, product=product, quantity=quantity)
            return stores_and_locations
        except OutOfStockError as e:
            if raise_on_none:
                raise e

    def warmup(
        self,
        *,
        products_generator: ProductsGenerator,
        locations: Literal["products", "random"],
    ):
        from eagle_trays.asrs import ASRS
        from eagle_trays.avsrs import AVSRS

        if locations == "products":
            for product in products_generator.products:
                for type_of_store, stores in self.stores.items():
                    if type_of_store is ASRS:
                        case_container = "pallet"
                    elif type_of_store is AVSRS:
                        case_container = "tray"
                    else:
                        raise ValueError

                    s_max = product.s_max[case_container]  # [cases]
                    n_pallet = math.ceil(s_max / product.case_per_pallet)

                    # if product.family in ("B", "C"):
                    #    n_pallet -= 1
                    n_pallet = max(1, n_pallet)

                    def iter_stores():
                        i = 0
                        while True:
                            try:
                                yield stores[i]
                                i += 1
                            except IndexError:
                                i = 0

                    if case_container == "pallet":
                        for _, store in zip(range(n_pallet), iter_stores()):
                            unit_load = Pallet.by_product(product=product)
                            location = store.first_available_location_for_warmup(unit_load=unit_load)
                            store.book_location(location=location, unit_load=unit_load)
                            location.put(unit_load=unit_load)

                            # aumentiamo l'on_hand
                            self.update_stock(
                                product=product,
                                case_container=case_container,
                                inventory="on_hand",
                                n_cases=unit_load.n_cases,
                            )
                    else:
                        for _ in range(n_pallet):
                            for _, store in zip(range(product.layers_per_pallet), iter_stores()):
                                unit_load = Pallet(
                                    Tray(
                                        product=product,
                                        n_cases=product.cases_per_layer,
                                    )
                                )
                                location = store.first_available_location_for_warmup(unit_load=unit_load)
                                store.book_location(location=location, unit_load=unit_load)
                                location.put(unit_load=unit_load)

                                # aumentiamo l'on_hand
                                self.update_stock(
                                    product=product,
                                    case_container=case_container,
                                    inventory="on_hand",
                                    n_cases=unit_load.n_cases,
                                )
        else:
            raise ValueError(f"Unknown locations warmup policy {locations}")



from __future__ import annotations

import random
from typing import Callable, Optional, TypeVar

import numpy as np

from simulatte.utils import Identifiable

T = TypeVar("T")
DistributionCallable = Callable[[], T]


class Product(metaclass=Identifiable):
    id: int

    def __init__(
        self,
        *,
        probability: float,
        family: str,
        cases_per_layer: int,
        layers_per_pallet: int,
        max_case_per_pallet: int,
        min_case_per_pallet: int,
        lp_enabled: bool,
    ) -> None:
        self.probability = probability
        self.family = family
        self.cases_per_layer = cases_per_layer
        self.layers_per_pallet = layers_per_pallet
        self.max_case_per_pallet = max_case_per_pallet
        self.min_case_per_pallet = min_case_per_pallet
        self.lp_enabled = lp_enabled
        self.s_max = {
            "pallet": 0,  # in cases
            "tray": 0,  # in cases
        }
        self.s_min = {
            "pallet": 0,  # in cases
            "tray": 0,  # in cases
        }

    def __repr__(self) -> str:
        return f"Product(id={self.id})"

    @property
    def case_per_pallet(self) -> int:
        return self.cases_per_layer * self.layers_per_pallet


class ProductsGenerator:
    def __init__(
        self,
        *,
        n_products: int,
        probabilities: Optional[DistributionCallable[list[float]]] = None,
        families: Optional[DistributionCallable[list[str]]] = None,
        cases_per_layers: Optional[DistributionCallable[int]] = None,
        layers_per_pallet: Optional[DistributionCallable[int]] = None,
        min_case_per_pallet: Optional[DistributionCallable[int]] = None,
        max_case_per_pallet: Optional[DistributionCallable[int]] = None,
        lp_enable: Optional[DistributionCallable[bool]] = None,
    ) -> None:

        self.probabilities: list[float] = (
            probabilities() if probabilities is not None else [1 / n_products for _ in range(n_products)]
        )
        self.families: list[str] = families() if families is not None else ["A"] * n_products
        self.cases_per_layers: DistributionCallable[int] = cases_per_layers or (lambda: 10)
        self.layers_per_pallet: DistributionCallable[int] = layers_per_pallet or (lambda: 4)
        self.min_case_per_pallet: DistributionCallable[int] = min_case_per_pallet or (lambda: 60)
        self.max_case_per_pallet: DistributionCallable[int] = max_case_per_pallet or (lambda: 60)
        self.lp_enable: DistributionCallable[bool] = lp_enable or (lambda: True)

        self._products: list[Product] | None = None

    @property
    def products(self) -> list[Product]:
        if self._products is None:
            self._products = [
                Product(
                    probability=probability,
                    family=family,
                    cases_per_layer=self.cases_per_layers(),
                    layers_per_pallet=self.layers_per_pallet(),
                    max_case_per_pallet=self.max_case_per_pallet(),
                    min_case_per_pallet=self.min_case_per_pallet(),
                    lp_enabled=self.lp_enable(),
                )
                for probability, family in zip(self.probabilities, self.families)
            ]
        return self._products

    def choose_one(
        self,
        *,
        exclude: set[Product] | None = None,
        fn: Optional[Callable[[list[Product]], Product]] = None,
    ) -> Product:
        def choose():
            if fn is not None:
                product = fn(self.products)
            else:
                product: Product = random.choices(self.products, weights=self.probabilities, k=1)[0]

            return product

        product = choose()

        if exclude is not None:
            while product in exclude:
                product = choose()

            exclude.add(product)

        return product

    def choose_some(
        self,
        *,
        n: int = 1,
        replace=False,
        fn: Optional[Callable[[list[Product]], Product]] = None,
    ) -> list[Product]:
        if n > len(self.products):
            raise ValueError("Cannot generate a sample larger than the population")

        if fn is not None:
            products = list(fn(self.products) for _ in range(n))
        else:
            products: list[Product] = np.random.choice(self.products, n, replace=replace, p=self.probabilities)

        return products



from __future__ import annotations

from functools import total_ordering
from typing import TYPE_CHECKING

from simulatte.events import LoggedEvent
from simulatte.unitload import Pallet
from simulatte.utils import Identifiable

if TYPE_CHECKING:
    from simulatte.ant import Ant
    from simulatte.picking_cell import PickingCell
    from simulatte.picking_cell.areas.position import Position
    from simulatte.requests import PickingRequest
    from simulatte.stores import WarehouseStore, WarehouseLocation


@total_ordering
class FeedingOperation(metaclass=Identifiable):
    """
    Represents a feeding operation assigned by the System to an ant.
    The ant is responsible for retrieving a unit load from a specific store, according to the
    picking request, and then to bring it to the assigned picking cell.
    """

    id: int

    def __init__(
        self,
        *,
        cell: PickingCell,
        ant: Ant,
        store: WarehouseStore,
        picking_request: PickingRequest,
        location,
        unit_load: Pallet,
    ) -> None:

        self.cell = cell
        self.relative_id = len(self.cell.feeding_operations)
        self.cell.feeding_operations.append(self)

        self.ant = ant
        self.store = store
        self.location: WarehouseLocation = location
        self.unit_load = unit_load
        self.picking_request = picking_request

        self.picking_request.feeding_operations.append(self)
        self.pre_unload_position: Position | None = None
        self.unload_position: Position | None = None

        self.status = {
            "arrived": False,
            "staging": False,
            "inside": False,
            "ready": False,
            "done": False,
        }
        self.ready = LoggedEvent(env=self.cell.system.env)

    def __repr__(self):
        return f"FeedingOperation{self.id}"

    def __lt__(self, other: FeedingOperation) -> bool:
        return self.id < other.id

    def __eq__(self, other: FeedingOperation) -> bool:
        return self.id == other.id

    @property
    def is_in_feeding_area(self) -> bool:
        return sum(self.status.values()) == 0

    def _check_status(self, *status_to_be_true) -> bool:
        for status in self.status:
            if status in status_to_be_true:
                if not self.status[status]:
                    return False
            else:
                if self.status[status]:
                    return False
        return True

    @property
    def is_in_front_of_staging_area(self) -> bool:
        return self._check_status("arrived")

    @property
    def is_inside_staging_area(self) -> bool:
        return self._check_status("arrived", "staging")

    @property
    def is_in_internal_area(self) -> bool:
        return self._check_status("arrived", "staging", "inside")

    @property
    def is_at_unload_position(self) -> bool:
        return self._check_status("arrived", "staging", "inside", "ready")

    @property
    def is_done(self) -> bool:
        return self._check_status("arrived", "staging", "inside", "ready", "done")

    def knock_staging_area(self) -> None:
        self.status["arrived"] = True

    def enter_staging_area(self) -> None:
        self.status["staging"] = True

    def enter_internal_area(self) -> None:
        self.status["inside"] = True

    def ready_for_unload(self) -> None:
        self.status["ready"] = True
        self.ready.succeed(value={"type": 0, "operation": self})

    def unloaded(self) -> None:
        self.status["done"] = True



from __future__ import annotations

from typing import TYPE_CHECKING

from simulatte.products import Product
from simulatte.stores import WarehouseStore

if TYPE_CHECKING:
    from simulatte.stores.warehouse_location.warehouse_location import WarehouseLocation


class LocationPolicy:
    def __call__(self, *, store: WarehouseStore, product: Product) -> WarehouseLocation | None:
        raise NotImplementedError


class ClosestLocationPolicy(LocationPolicy):
    def __call__(self, *, store: WarehouseStore, product: Product) -> WarehouseLocation | None:

        possible_locations = [
            location
            for location in store.locations
            if location.product is None  # vanno bene locazioni completamente vuote
            or location.product == product  # consideriamo solo locazioni che gestiscono o gestiranno il prodotto
            # consideriamo solo le locazioni che hanno disponibilità (fisica o futura)
            and location.n_unit_loads + len(location.future_unit_loads) < location.depth
        ]

        aff = [location.affinity(product=product) for location in possible_locations]
        if all(a == float("inf") for a in aff):
            raise ValueError

        sorted_locations = sorted(
            possible_locations,
            key=lambda location: location.affinity(product=product),
        )
        if sorted_locations:
            selected_location = sorted_locations[0]
            return selected_location



from .operation import InputOperation, OutputOperation
from .warehouse_location import distance
from .warehouse_location.warehouse_location import (
    WarehouseLocation,
    WarehouseLocationSide,
)
from .warehouse_store import WarehouseStore



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from simulatte.products import Product

from simulatte.requests import LayerRequest, PalletRequest, ProductRequest


def fixed_sequence(products: list[Product], n_pallet_requests: int, n_layers: int) -> Iterable[PalletRequest]:
    i = -1
    sequence = [[i := i + 1 for _ in range(n_layers)] for _ in range(n_pallet_requests)]

    for r in sequence:
        yield PalletRequest(*[LayerRequest(ProductRequest(products[i], products[i].cases_per_layer)) for i in r])



from __future__ import annotations

from typing import TYPE_CHECKING

from simulatte.unitload import CaseContainer

if TYPE_CHECKING:
    from simulatte.products import Product


class Tray(CaseContainer):
    def __init__(self, *, product: Product, n_cases: int, exceed=False) -> None:
        if n_cases > product.cases_per_layer and not exceed:
            raise ValueError(
                f"A Tray cannot hold n_cases={n_cases} [product.cases_per_layer={product.cases_per_layer}]"
            )
        self.product = product
        self.n_cases = n_cases

    def __repr__(self) -> str:
        return f"Tray(product={self.product}, n_cases={self.n_cases})"



from __future__ import annotations

import pprint
from datetime import datetime
from typing import TYPE_CHECKING, Optional, TypedDict

import pandas as pd

import simulatte

if TYPE_CHECKING:
    from simulatte.picking_cell import FeedingOperation, PickingCell


class EventPayload(TypedDict, total=False):
    time: float
    cell: PickingCell
    event: str
    type: int
    operation: Optional[FeedingOperation]


class Logger(metaclass=simulatte.utils.Singleton):
    """
    Used to register events during a simulation.
    """

    def __init__(self) -> None:
        self.logs: list[EventPayload] = []

    def log(self, *, payload: EventPayload) -> None:
        self.logs.append(payload)

    def reset(self) -> None:
        self.logs = []

    def export(self, *, filename: str | None = None) -> None:
        df = pd.DataFrame(self.logs)
        filename = filename or f"{datetime.now()}.csv"
        df.to_csv(f"./export_data/{filename}.csv")

    def __str__(self) -> str:
        return pprint.pformat(self.logs)



from __future__ import annotations

from simulatte.operations import FeedingOperation

from .observable_area import ObservableArea


class StagingArea(ObservableArea[FeedingOperation]):
    """
    Represent the logical area inside a picking cell where ants wait to be processed.
    """

    def append(self, feeding_operation: FeedingOperation) -> None:
        super().append(feeding_operation)
        feeding_operation.enter_staging_area()



from __future__ import annotations

from typing import TYPE_CHECKING

from .base_observer import Observer
from ..areas import FeedingArea

if TYPE_CHECKING:
    from simulatte.requests import Request


class FeedingObserver(Observer[FeedingArea]):
    def next(self) -> Request:
        return self.observable_area.cell.picking_requests_queue.popleft()

    def _main_process(self) -> None:
        self.system.start_feeding_operation(cell=self.cell)



from __future__ import annotations

from simpy.resources.base import BoundClass
from simpy.resources.resource import Release, PriorityRequest, PriorityResource

from simulatte.utils import Priority


class MonitoredRequest(PriorityRequest):
    """
    Request for the MonitoredResource.
    """

    resource: MonitoredResource

    def __init__(self, resource: MonitoredResource, *, priority=Priority.NORMAL, preempt=False) -> None:
        super().__init__(resource, priority=priority, preempt=preempt)

    def __enter__(self):
        self.start_time = self.env.now
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        end_time = self.env.now
        self.resource.worked_time += end_time - self.start_time
        return super().__exit__(exc_type, exc_val, exc_tb)


class MonitoredResource(PriorityResource):
    """
    PriorityResource with monitoring capabilities.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._worked_time = 0
        self._saturation_history = []

    @property
    def worked_time(self) -> int:
        return self._worked_time

    @worked_time.setter
    def worked_time(self, value: int) -> None:
        self._worked_time = value
        self._saturation_history.append((self._env.now, self.saturation))

    @property
    def idle_time(self) -> int:
        return self._env.now - self._worked_time

    @property
    def saturation(self) -> float:
        return self._worked_time / self._env.now

    request = BoundClass(MonitoredRequest)
    release = BoundClass(Release)



from __future__ import annotations

from typing import TYPE_CHECKING

from simulatte.operations import FeedingOperation

from .observable_area import ObservableArea
from .position import Position

if TYPE_CHECKING:
    from simulatte.picking_cell import PickingCell


class InternalArea(ObservableArea[FeedingOperation]):
    """
    Represent the area inside a picking cell, where ants are placed to be unloaded.
    Manage both unloading and pre-unloading positions.
    """

    def __init__(self, *, cell: PickingCell, capacity: int, pre_unload: bool = False) -> None:
        super().__init__(cell=cell, capacity=capacity)

        if pre_unload:
            self.unload_positions = tuple(
                Position(name=f"UnloadPosition{i}", env=cell.system.env, capacity=1) for i in range(capacity // 2)
            )
            self.pre_unload_positions = tuple(
                Position(name=f"PreUnloadPosition{i}", env=cell.system.env, capacity=1) for i in range(capacity // 2)
            )
        else:
            self.unload_positions = tuple(
                Position(name=f"UnloadPosition{i}", env=cell.system.env, capacity=1) for i in range(capacity)
            )
            self.pre_unload_positions = tuple()

    def append(self, feeding_operation: FeedingOperation) -> None:
        super().append(feeding_operation)
        feeding_operation.enter_internal_area()

    def remove(self, feeding_operation: FeedingOperation) -> None:
        super().remove(feeding_operation)
        feeding_operation.unloaded()



from simulatte.logger.logger import EventPayload
from simulatte.operations import FeedingOperation

from ..areas import StagingArea
from .base_observer import Observer


class StagingObserver(Observer[StagingArea]):
    def next(self) -> FeedingOperation | None:
        """
        Select the FeedingOperation allowed to exit the FeedingArea and enter the StagingArea.
        """
        feeding_operations = (
            feeding_operation
            for feeding_operation in self.observable_area.cell.feeding_area
            if feeding_operation.is_in_front_of_staging_area
        )
        return min(feeding_operations, default=None)

    def _can_enter(self, *, feeding_operation: FeedingOperation) -> bool:
        """
        Check if the feeding operation can enter the staging area.

        The feeding operation can enter the staging area if the staging area is not full and the feeding operation is in
        front of the staging area.
        """

        is_first_ever_feeding_operation = self.cell.internal_area.last_entered is None and feeding_operation is min(
            self.cell.feeding_operations
        )
        is_next_useful_feeding_operation = (
            self.cell.internal_area.last_entered is not None
            and feeding_operation.relative_id == self.cell.internal_area.last_entered.relative_id + 1
        )

        return is_first_ever_feeding_operation or is_next_useful_feeding_operation

    def _main_process(self):
        """
        Manage the entering processes of a picking cell.

        Called when the StagingArea signal event is triggered (see WMS._feeding_process).

        It checks if the feeding area is not empty and the staging area is not full.

        The procedure follows the following steps:
        1. Removes the feeding operation from the feeding area.
        2. Moves the feeding operation to the staging area.
        3. Updates the status of the feeding operation.
        4. Signal the feeding area that a feeding operation has been removed.
        5. Signal the internal area that a new feeding operation has entered the staging area.
        """

        if (
            not self.cell.feeding_area.is_empty
            and not self.cell.staging_area.is_full
            and (feeding_operation := self.next()) is not None
        ):
            if self._can_enter(feeding_operation=feeding_operation):
                # Remove the FeedingOperation from the FeedingArea
                self.cell.feeding_area.remove(feeding_operation)

                # The FeedingOperation enters the StagingArea
                self.cell.staging_area.append(feeding_operation)

                feeding_operation.ant.enter_staging_area()
                feeding_operation.ant.move_to(system=self.system, location=self.cell.staging_location)

                # Trigger the InternalArea signal event since a new FeedingOperation is available in the StagingArea
                payload = EventPayload(event="ACTIVATING INTERNAL AREA SIGNAL", type=0)
                self.cell.internal_area.trigger_signal_event(payload=payload)
                return



from __future__ import annotations

from enum import Enum, auto
from typing import TYPE_CHECKING, Iterable

import matplotlib.pyplot as plt
from simpy import PriorityResource

from simulatte.location import Location
from simulatte.typings import ProcessGenerator
from simulatte.unitload import CaseContainer
from simulatte.utils import Identifiable, as_process

if TYPE_CHECKING:
    from simpy import Environment

    from simulatte.system import System


class AntStatus(Enum):
    """
    The status of an ant.
    """

    IDLE = auto()
    TRAVELING_UNLOADED = auto()
    TRAVELING_LOADED = auto()
    WAITING_UNLOADED = auto()
    WAITING_LOADED = auto()


class AntRestLocation(Location):
    def __init__(self):
        super().__init__(name="AntRestLocation")


ant_rest_location = AntRestLocation()


class AntMission:
    def __init__(self, ant: Ant, start_location: Location, end_location: Location):
        self.ant = ant
        self.start_location = start_location
        self.start_time = ant.env.now
        self.end_location = end_location
        self.end_time = None

    def end(self):
        self.end_time = self.ant.env.now
        self.ant.mission_logs.append(self)


class Ant(PriorityResource, metaclass=Identifiable):
    """
    Represent a generic Ant.

    The ant time is divided into 3 phases:
    - Traveling (whether loaded or unloaded)
    - Waiting (whether loaded or unloaded)
    - Idling (unloaded, no mission assigned)

    The mission time is the sum of traveling and waiting times.

    +-----------------+-----------------+-----------------+
    |     TRAVEL      |     WAITING     |       IDLE      |
    |      TIME       |      TIME       |       TIME      |
    +-----------------+-----------------+-----------------+
    |            MISSION TIME           |
    +-----------------+-----------------+
    """

    id: int

    def __init__(self, env: Environment, kind: str, load_timeout=0, unload_timeout=0) -> None:
        super().__init__(env, capacity=1)
        self.env = env
        self.kind = kind
        self._case_container: CaseContainer | None = None
        self.load_timeout = load_timeout
        self.unload_timeout = unload_timeout
        self._status = AntStatus.IDLE
        self.current_location: Location = ant_rest_location
        self._travel_time = 0
        self._mission_history: list[float] = []
        self.mission_logs: list[AntMission] = []

        self.loading_waiting_times = []
        self.loading_waiting_time_start: float | None = None

        self._waiting_to_enter_staging_area: float | None = None
        self.feeding_area_waiting_times = []

        self._waiting_to_enter_internal_area: float | None = None
        self.staging_area_waiting_times = []

        self._waiting_to_be_unloaded: float | None = None
        self.unloading_waiting_times = []

        self._waiting_picking_to_end: float | None = None
        self.picking_waiting_times = []

    @property
    def status(self) -> AntStatus:
        return self._status

    @status.setter
    def status(self, value: AntStatus) -> None:
        self._status = value

    @property
    def idle_time(self) -> float:
        return self.env.now - self.mission_time

    @property
    def saturation(self) -> float:
        return self.mission_time / self.env.now

    @property
    def missions(self) -> Iterable[tuple[float, float]]:
        for start, end in zip(self._mission_history[::2], self._mission_history[1::2]):
            yield start, end

    @property
    def mission_time(self) -> float:
        return sum(end - start for start, end in self.missions)

    @property
    def waiting_time(self) -> float:
        return self.mission_time - self._travel_time

    @property
    def unit_load(self) -> CaseContainer | None:
        return self._case_container

    @unit_load.setter
    def unit_load(self, value: CaseContainer | None) -> None:
        if self._case_container is not None and value is not None:
            raise RuntimeError(f"Ant [{self.id}] cannot carry two unit loads at the same time.")
        self._case_container = value

    def idle(self) -> None:
        """Set the ant to idle status"""
        self.status = AntStatus.IDLE

    def waiting_to_be_loaded(self) -> None:
        """Set the ant to waiting status"""
        self.status = AntStatus.WAITING_UNLOADED
        self.loading_waiting_time_start = self.env.now

    def waiting_to_be_unloaded(self) -> None:
        """Set the ant to waiting status"""
        self.status = AntStatus.WAITING_LOADED

    def waiting_to_enter_staging_area(self) -> None:
        self._waiting_to_enter_staging_area = self.env.now

    def enter_staging_area(self) -> None:
        self.feeding_area_waiting_times.append(self.env.now - self._waiting_to_enter_staging_area)
        self._waiting_to_enter_staging_area = None
        self._waiting_to_enter_internal_area = self.env.now

    def enter_internal_area(self):
        self.staging_area_waiting_times.append(self.env.now - self._waiting_to_enter_internal_area)
        self._waiting_to_enter_internal_area = None
        self._waiting_to_be_unloaded = self.env.now

    def picking_begins(self):
        if self._waiting_to_be_unloaded is not None:
            self.unloading_waiting_times.append(self.env.now - self._waiting_to_be_unloaded)
        self._waiting_to_be_unloaded = None
        self._waiting_picking_to_end = self.env.now

    def picking_ends(self):
        self.picking_waiting_times.append(self.env.now - self._waiting_picking_to_end)
        self._waiting_picking_to_end = None

    @as_process
    def load(self, *, unit_load: CaseContainer) -> ProcessGenerator:
        self.unit_load = unit_load
        yield self.env.timeout(self.load_timeout)
        if self.loading_waiting_time_start is not None:
            self.loading_waiting_times.append(self.env.now - self.loading_waiting_time_start)
        self.loading_waiting_time_start = None
        self.status = AntStatus.WAITING_LOADED

    @as_process
    def unload(self) -> ProcessGenerator:
        if self.unit_load is None:
            raise ValueError("Ant cannot unload non-existent unit load.")
        yield self.env.timeout(self.unload_timeout)
        self.unit_load = None
        self.status = AntStatus.WAITING_LOADED

    @as_process
    def move_to(self, *, system: System, location: Location):
        mission = AntMission(ant=self, start_location=self.current_location, end_location=location)

        timeout = system.distance(self.current_location, location).as_time

        if self.unit_load is not None:
            self.status = AntStatus.TRAVELING_LOADED
        else:
            self.status = AntStatus.TRAVELING_UNLOADED

        yield self.env.timeout(timeout)
        self._travel_time += timeout

        mission.end()

        self.current_location = location

        if self.unit_load is not None:
            self.status = AntStatus.WAITING_LOADED
        else:
            self.status = AntStatus.WAITING_UNLOADED

    def release_current(self):
        """Release the current request the ant is taking care of"""
        if len(self.users) == 0:
            raise ValueError("Ant cannot release non-existent request.")
        self.release(self.users[0])

    def mission_started(self) -> None:
        self._mission_history.append(self.env.now)
        self.status = AntStatus.WAITING_UNLOADED

    def mission_ended(self) -> None:
        self._mission_history.append(self.env.now)
        self.status = AntStatus.IDLE

    def plot(self) -> None:
        plt.plot([(end - start) / 60 for start, end in self.missions], "o-")
        plt.title(f"Ant [{self.id}] mission duration [min]")
        plt.show()



from __future__ import annotations

from typing import Iterable

from simpy import Process
from simpy.resources.resource import PriorityRequest

from simulatte.distance import Distance
from simulatte.environment import Environment
from simulatte.location import Location
from simulatte.logger import Logger
from simulatte.operations import FeedingOperation
from simulatte.picking_cell import PickingCell
from simulatte.requests import PalletRequest
from simulatte.stores import WarehouseStore
from simulatte.system.managers import AntsManager, StoresManager
from simulatte.system.managers.cells_manager import CellsManager
from simulatte.system.managers.distance_manager import DistanceManager


class System:
    feeding_operations: list[FeedingOperation]
    ants_manager: AntsManager
    stores_manager: StoresManager
    cells_manager: CellsManager
    distance_manager: DistanceManager

    stores: Iterable[WarehouseStore]
    locations: Iterable[Location]

    def __init__(self, env: Environment) -> None:
        self.env = env
        self.cells_manager = CellsManager(system=self)
        self.distance_manager = DistanceManager(system=self, DistanceClass=Distance)

        self.logger = Logger()

    def distance(self, from_: Location, to: Location) -> Distance:
        raise NotImplementedError

    def get_store_by_cell(self, *, cell: PickingCell | None = None) -> WarehouseStore:
        raise NotImplementedError

    @staticmethod
    def assign_to_cell(*, cell: PickingCell, pallet_request: PalletRequest) -> Process:
        raise NotImplementedError

    def retrieve_from_cell(self, *, cell: PickingCell, pallet_request: PalletRequest) -> Process:
        raise NotImplementedError

    def start_feeding_operation(self, *, cell: PickingCell) -> None:
        raise NotImplementedError

    def feed(self, *, feeding_operation: FeedingOperation, ant_request: PriorityRequest):
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING

from simulatte.events.logged_event import LoggedEvent


if TYPE_CHECKING:
    from simulatte.system import System
    from simulatte.logger.logger import EventPayload


class Observable:
    """
    Implements the observer/observable pattern.

    An observable is observed by an observer.
    The observer uses the observable signal event to act accordingly.
    """

    def __init__(self, *, system: System) -> None:
        self.system = system
        self.signal_event = self._init_signal_event()

    def _init_signal_event(self) -> LoggedEvent:
        """
        Initialize the observable signal event.
        """
        return LoggedEvent(env=self.system.env)

    @property
    def _logged_event_title(self) -> str:
        return self.__class__.__name__

    def trigger_signal_event(self, *, payload: EventPayload) -> None:
        """
        Trigger the observer signal event and then reset it.
        """
        payload["time"] = self.system.env.now
        if hasattr(self, "cell"):
            payload["cell"] = self.cell.name

        self.signal_event.succeed(value=payload)
        self.signal_event = self._init_signal_event()



from __future__ import annotations

from simpy.events import Event

from simulatte.logger.logger import EventPayload, Logger


class LoggedEvent(Event):
    def succeed(self, value: EventPayload | None = None):
        self.timestamp = self.env.now
        payload = value or {}
        if payload:
            Logger().log(payload=payload)
        return super().succeed(value=payload.get("value"))



from itertools import count


class Identifiable(type):

    id: int

    def __call__(cls, *args, **kwargs):
        if not hasattr(cls, "_id_iter"):
            cls._id_iter = count()
            cls._instances = {}
        _instance = super(Identifiable, cls).__call__(*args, **kwargs)
        _instance.id = next(cls._id_iter)
        cls._instances[_instance.id] = _instance
        return _instance



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from simpy import Environment

from simpy.core import BoundClass
from simpy.resources.base import BaseResource

from .multi_store_get import MultiStoreGet
from .multi_store_put import MultiStorePut


class MultiStore(BaseResource):
    """
    The MultiStore works like a 'simpy.resources.store.Store',
    but allows the storage and retrieval of more elements at the same time.

    If the user tries to get more items than the ones available,
    the MultiStore will return all the available items.
    """

    put = BoundClass(MultiStorePut)

    get = BoundClass(MultiStoreGet)

    def __init__(self, env: Environment, capacity: int = float("inf")) -> None:
        if capacity <= 0:
            raise ValueError("capacity must be > 0.")

        super().__init__(env, capacity)
        self.items = []

    def _do_put(self, event: MultiStorePut) -> None:
        """
        Put a collection of items into the store, if there is enough space.
        The items to be stored are passed in the 'event.items' attribute.
        """
        if self.level + len(event.items) <= self._capacity:  # enough space
            # store the items
            self.items.extend(event.items)
            # notify the event
            event.succeed()
        return None

    def _do_get(self, event: MultiStoreGet) -> None:
        """
        Get a collection of items from the store.

        If there are fewer items than requested, the MultiStore will return
        all the available items.
        """
        if self.items:  # at least one item is available
            # return the first n items
            to_return, self.items = self.items[: event.n], self.items[event.n :]
            # notify the event
            event.succeed(to_return)

    @property
    def level(self) -> int:
        """
        The number of items currently in the store.
        """
        return len(self.items)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Generic, TypeVar


from ...unitload import CaseContainer
from ...utils import Identifiable
from .exceptions import IncompatibleUnitLoad, LocationBusy, LocationEmpty
from .physical_position import PhysicalPosition

if TYPE_CHECKING:
    from simulatte.products import Product
    from .. import WarehouseStore


class WarehouseLocationSide(Enum):
    LEFT = "left"
    RIGHT = "right"
    ORIGIN = "origin"


T = TypeVar("T", bound=CaseContainer)


class WarehouseLocation(Generic[T], metaclass=Identifiable):
    """
    Warehouse physical storage location, where the unit loads are stored.
    """

    id: int

    def __init__(
        self,
        *,
        store: WarehouseStore[T],
        x: int,
        y: int,
        side: WarehouseLocationSide,
        depth: int = 2,
        width: float,
        height: float,
    ) -> None:
        """
        :param x: The discrete x-coordinate of the location (x-axis).
        :param y: The discrete y-coordinate of the location (y-axis).
        :param side: 'left' or 'rigth'.
        :param depth: The depth of the storage location.
        """
        self.store = store
        self.x = x
        self.y = y

        if side not in WarehouseLocationSide:
            raise ValueError("Side must be a value of WarehouseLocationSide")
        self.side = side.value

        if depth not in (1, 2):
            raise ValueError("The depth of the location must be positive and cannot be greater than 2.")
        self.depth = depth

        self.width = width
        self.height = height

        self.first_position = PhysicalPosition()
        self.second_position = PhysicalPosition()
        self.future_unit_loads: list[T] = []  # Unit loads that will be stored in the future
        self.booked_pickups: list[T] = []  # Unit loads that will be picked up in the future

    def __repr__(self) -> str:
        return (
            f"Location ({self.x}, {self.y}, {self.side}) reserved for {self.product} "
            f"[{len(self.future_unit_loads)}, {self.n_unit_loads}, "
            f"{self.is_empty}, {self.is_half_full}, {self.is_full}]"
        )

    @property
    def coordinates(self) -> tuple[float, float]:
        """
        The location coordinates in meters.
        """
        return self.x * self.width, self.y * self.height

    def check_product_compatibility(self, unit_load: T) -> bool:
        if unit_load.product != self.first_available_unit_load.product:
            raise IncompatibleUnitLoad(unit_load, self)
        return True

    def book_pickup(self, unit_load: T) -> None:
        if unit_load not in (
            self.first_position.unit_load,
            self.second_position.unit_load,
        ):
            raise RuntimeError("Unit load not available for pickup")

        if self.fully_booked:
            raise ValueError(f"Cannot book more than {self.depth} pickups")

        if unit_load in self.booked_pickups:
            raise ValueError("Unit load already booked")

        self.booked_pickups.append(unit_load)

    @property
    def fully_booked(self) -> bool:
        # first_position = self.first_position
        # second_position = self.second_position
        #
        # if first_position.free and second_position.free:
        #     return False
        #
        # for position in (self.first_position, self.second_position):
        #     if position.unit_load is not None:
        #         if position.unit_load not in self.booked_pickups:
        #             return False
        # return True

        return len(self.booked_pickups) == self.depth

    def freeze(self, *, unit_load: T) -> None:
        """
        Freeze the location for a certain unit load.

        If the location is not empty, first it checks that the unit load is compatible
        (same product as the one already stored).
        """
        if self.is_empty:
            if len(self.future_unit_loads) == self.depth:
                raise ValueError(
                    f"Cannot freeze a location with empty positions, but with {self.depth} future unit loads"
                )
        elif self.is_half_full:
            if len(self.future_unit_loads) >= 1:
                raise ValueError("Cannot freeze a location with one busy position, and one future unit load")
        else:
            raise ValueError("Cannot freeze a location with two busy positions")

        if not self.is_empty:
            self.check_product_compatibility(unit_load)

        self.future_unit_loads.append(unit_load)

    def unfreeze(self, unit_load: T) -> None:
        if unit_load not in self.future_unit_loads:
            raise ValueError("Unit load not found in the future unit loads")

        self.future_unit_loads.remove(unit_load)

    @property
    def physically_available_product(self) -> Product | None:
        """
        Returns the product that is stored in the location.
        A product can be stored in the location in the following cases:
        - The location is empty.
        - The location is half full and the product is compatible with the one already stored.
        """
        if self.is_empty:
            return None
        return self.first_available_unit_load.product

    @property
    def product(self) -> Product | None:
        """
        Returns the product associated with the location.
        A product is associated with the location in the following cases:
        - The location is not empty.
        - The location is empty, but it is frozen for a certain incoming product.
        """
        if self.is_empty:
            if len(self.future_unit_loads) == 0:
                # No unit load is stored in the location and no unit load is planned to be stored
                return None
            else:
                # No unit load is stored in the location but a unit load is planned to be stored
                return self.future_unit_loads[0].product
        else:
            return self.first_available_unit_load.product

    @property
    def n_cases(self) -> int:
        """
        Returns the number of cases stored in the location.
        """
        return (
            self.first_position.n_cases
            + self.second_position.n_cases
            + sum(unit_load.n_cases for unit_load in self.future_unit_loads)
        )

    def deals_with_products(self, product: Product) -> bool:
        return self.product == product

    @property
    def is_empty(self) -> bool:
        if self.depth == 2:
            return self.first_position.free and self.second_position.free
        return self.first_position.free

    @property
    def is_half_full(self) -> bool:
        return self.second_position.busy and self.first_position.free

    @property
    def is_full(self) -> bool:
        if self.depth == 2:
            return self.first_position.busy and self.second_position.busy
        return self.first_position.busy

    @property
    def n_unit_loads(self) -> int:
        return int(self.first_position.busy) + int(self.second_position.busy)

    @property
    def first_available_unit_load(self) -> T:
        """
        Returns the first available unit load.
        If the first position is busy, the unit load in second position is not available.
        """
        if self.is_empty:
            raise LocationEmpty(self)

        if self.is_full:
            return self.first_position.unit_load

        if self.is_half_full:
            return self.second_position.unit_load

    @property
    def first_available_position(self) -> PhysicalPosition:
        """
        Returns the first available unit load.
        If the first position is busy, the unit load in second position is not available.
        """
        if self.is_empty:
            raise LocationEmpty(self)

        if self.is_full:
            return self.first_position

        if self.is_half_full:
            return self.second_position

    def put(self, unit_load: T) -> None:
        """
        Stores a unit load into the location.

        If the location is empty, the unit load will be stored into the second position.

        If the location is half full, the unit load will be stored into the first position,
        if the unit load is compatible (must contain the same product already stored).

        If the location is full, an exception will be raised.
        """
        if self.is_empty:
            # TODO: gestire singola profondità
            physical_position = self.second_position
        elif self.is_half_full:
            existing_product = self.second_position.unit_load.product
            if unit_load.product != existing_product:
                raise IncompatibleUnitLoad(unit_load, self)
            physical_position = self.first_position
        else:
            raise LocationBusy(self)

        physical_position.put(unit_load=unit_load)
        unit_load.location = self
        self.unfreeze(unit_load=unit_load)

    def get(self, unit_load) -> T:
        if len(self.booked_pickups) == 0:
            raise ValueError("Cannot get a unit load without booking it first")

        if self.is_empty:
            raise LocationEmpty(self)

        if unit_load is self.first_position.unit_load:
            physical_position = self.first_position
        elif unit_load is self.second_position.unit_load:
            # TODO: swap positions
            self.first_position.unit_load, self.second_position.unit_load = (
                self.second_position.unit_load,
                self.first_position.unit_load,
            )
            physical_position = self.first_position
        else:
            raise ValueError

        unit_load = physical_position.get()
        try:
            self.booked_pickups.remove(unit_load)
        except ValueError:
            raise ValueError("Cannot get a unit load without booking it first")
        unit_load.location = None
        return unit_load

    def affinity(self, product: Product) -> float:
        """
        Returns the affinity of the location to a certain product.
        The lower the number, the better is the location for the product.
        """

        if len(self.future_unit_loads) + self.n_unit_loads == self.depth:
            # la locazione è completamente occupata
            return float("inf")

        product_stored = self.product

        if product_stored == product:
            # la locazione contiene/conterrà il prodotto
            return 0
        if product_stored is None:
            # la locazione è vuota
            return 1

        if product_stored != product:
            # la locazione contiene un prodotto diverso
            return float("inf")



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter

if TYPE_CHECKING:
    from simulatte.picking_cell import PickingCell


T = TypeVar("T")


class Area(list, Generic[T]):
    """
    Implement a virtual area of interest for a picking cell.
    Extends list to add finite capacity.
    """

    def __init__(self, *, cell: PickingCell, capacity: int = float("inf")) -> None:
        super().__init__()
        self.env = cell.system.env
        self.cell = cell
        self.capacity = capacity
        self._history = []
        self.last_entered: T = None
        self.last_gone: T = None

    def pop(self, index: int = -1) -> T:
        item = super().pop(index)
        self.last_gone = item
        self._history.append((self.env.now, len(self)))
        return item

    def remove(self, item: T) -> None:
        super().remove(item)
        self.last_gone = item
        self._history.append((self.env.now, len(self)))

    @property
    def is_full(self) -> bool:
        return len(self) >= self.capacity

    @property
    def is_empty(self) -> bool:
        return len(self) == 0

    @property
    def free_space(self) -> int:
        return self.capacity - len(self)

    def append(self, item: T, exceed=False) -> None:
        if self.is_full and not exceed:
            raise ValueError("Area is full.")

        super().append(item)
        self.last_entered = item
        self._history.append((self.env.now, len(self)))

    def plot(self):
        x = [t / 60 / 60 for t, _ in self._history]
        y = [s for _, s in self._history]
        plt.plot(x, y)
        plt.yticks(range(0, max(y) + 1))
        plt.xlabel("Time [h]")
        plt.ylabel(f"Queue [#items]")
        plt.title(f"{self.__class__.__name__} queue")
        plt.show()



from __future__ import annotations


class Seconds:
    multiplier = 1
    value: int

    def __init__(self, value: int | Seconds = 0) -> None:
        if not isinstance(value, int):
            value = value.value
        self.value = value

    def __repr__(self) -> str:
        return f"{self.value} {self.__class__.__name__}"

    def __mul__(self, other) -> Seconds:
        prev_cls = self.__class__.mro()[1]
        if prev_cls in (Seconds, object):
            return other * self.multiplier
        return other * self.multiplier * prev_cls()

    def __rmul__(self, other) -> Seconds:
        return self.__mul__(other)


class Minutes(Seconds):
    multiplier = 60


class Hours(Minutes):
    multiplier = 60


class Days(Hours):
    multiplier = 24



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from simpy import Process
    from typing import Generator, Callable


def make_process(generator: Generator) -> Process:
    from simulatte.environment import Environment

    return Environment().process(generator)


def as_process(f: Callable[..., Generator]) -> Callable[..., Process]:
    @wraps(f)
    def wrapper(*args, **kwargs) -> Process:
        return make_process(f(*args, **kwargs))

    return wrapper



from .filter_multi_store import FilterMultiStore
from .hash_store import HashStore
from .multi_store import MultiStore
from .queue import PriorityQueue, PriorityItem
from .sequential_multi_store import SequentialMultiStore
from .sequential_store import SequentialStore



from __future__ import annotations

from typing import TYPE_CHECKING

from .exceptions import PhysicalPositionBusy, PhysicalPositionEmpty

if TYPE_CHECKING:
    from simulatte.unitload import Pallet


class PhysicalPosition:
    """
    Represent the physical position within a WarehouseLocation.
    """

    def __init__(self, unit_load: Pallet | None = None) -> None:
        self.unit_load = unit_load

    @property
    def n_cases(self) -> int:
        """
        Return the number of cases in the physical position.
        """
        if self.unit_load is not None:
            return self.unit_load.n_cases
        return 0

    def put(self, *, unit_load: Pallet) -> None:
        """
        Load a unit load into the physical position.
        """
        if self.busy:
            raise PhysicalPositionBusy(self)
        self.unit_load = unit_load

    def get(self) -> Pallet:
        """
        Empty the physical position and return the unit load.
        Raises PhysicalPositionEmpty if the physical position is free.
        """
        if self.free:
            raise PhysicalPositionEmpty(self)
        unit_load = self.unit_load
        self.unit_load = None
        return unit_load

    @property
    def free(self) -> bool:
        """
        Return True if no unit load is in the physical position.
        """
        return self.unit_load is None

    @property
    def busy(self) -> bool:
        """
        Return True if a unit load is in the physical position.
        """
        return not self.free



from typing import Type

from simpy.resources.store import Store

import simulatte
from simulatte.unitload import PaperSheet, WoodBoard


class EOQBuffer:
    """
    An instance of this class represents a buffer which is managed
    using a reorder level inventory policy.
    """

    def __init__(
        self,
        *,
        items_type: Type[PaperSheet | WoodBoard],
        reorder_level: int,
        eoq: int,
        get_time: int,
        put_time: int,
        capacity: int = float("inf"),
        init: int = 0,
    ):
        self.env = simulatte.Environment()
        self.items_type = items_type
        self.reorder_level = reorder_level
        self.eoq = eoq
        self.get_time = get_time
        self.put_time = put_time

        self.store = Store(self.env, capacity=capacity)

        if init > capacity:
            raise ValueError("Initial value is greater than capacity")

        for _ in range(init):
            self.store.items.append(items_type())

    @property
    def level(self) -> int:
        return len(self.store.items)

    @property
    def capacity(self) -> int:
        return self.store.capacity

    @property
    def need_refill(self) -> bool:
        """
        This property returns True if the buffer need to be filled,
        and False otherwise.
        """
        return self.level < self.reorder_level

    @simulatte.as_process
    def get(self):
        """
        A process to retrieve an item from the buffer.
        """
        item = yield self.store.get()
        yield self.env.timeout(self.get_time)
        return item

    @simulatte.as_process
    def put(self, *, items):
        """
        A process to store some items into the buffer.
        """
        for item in items:
            yield self.env.timeout(self.put_time)
            yield self.store.put(item)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Generic, TypeVar

import simulatte
from simulatte.stores import InputOperation, WarehouseLocation, WarehouseLocationSide
from .warehouse_location import distance

from ..unitload import CaseContainer, Pallet, Tray
from ..utils import Identifiable, as_process

if TYPE_CHECKING:
    from simpy.resources.store import Store

    from simulatte.ant import Ant
    from simulatte.operations import FeedingOperation
    from simulatte.products import Product
    from simulatte.service_point import ServicePoint
    from simulatte.simpy_extension import MultiStore, SequentialStore
    from ..system.managers import StoresManager


T = TypeVar("T", bound=CaseContainer)


class WarehouseStore(Generic[T], metaclass=Identifiable):
    id: int

    input_conveyor: Store | MultiStore
    output_conveyor: SequentialStore

    input_service_point: ServicePoint

    def __init__(
        self,
        *,
        stores_manager: StoresManager,
        n_positions: int = 20,
        n_floors: int = 8,
        location_width: float = 1,
        location_height: float = 1,
        depth: int = 2,
        load_time: int = 20,
        conveyor_capacity: int = 5,
    ):
        self.env = simulatte.Environment()
        self.stores_manager = stores_manager
        self.input_location = simulatte.location.Location(name=f"{self.__class__.__name__} Input")
        self.output_location = simulatte.location.Location(name=f"{self.__class__.__name__} Output")
        self.location_width = location_width
        self.location_height = location_height
        self.depth = depth
        self.n_positions = n_positions
        self.n_floors = n_floors
        self.load_time = load_time
        self.conveyor_capacity = conveyor_capacity

        self.queue_stats = []

        self._location_origin = WarehouseLocation(
            store=self,
            x=0,
            y=0,
            width=self.location_width,
            height=self.location_height,
            depth=self.depth,
            side=WarehouseLocationSide.ORIGIN,
        )
        self._locations = tuple(
            sorted(
                (
                    WarehouseLocation(
                        store=self,
                        x=x,
                        y=y,
                        side=side,
                        depth=self.depth,
                        width=self.location_width,
                        height=self.location_height,
                    )
                    for x in range(self.n_positions)
                    for y in range(self.n_floors)
                    for side in [
                        WarehouseLocationSide.LEFT,
                        WarehouseLocationSide.RIGHT,
                    ]
                ),
                key=lambda location: distance.euclidean(location, self._location_origin),
            )
        )

        self._input_operations = []
        self._replenishment_processes = defaultdict(list)
        self._product_location_map = defaultdict(set)
        self._saturation_history = []

    @property
    def locations(self):
        return self._locations

    @property
    def n_locations(self) -> int:
        return len(self.locations)

    @property
    def name(self) -> str:
        return f"{self.__class__.__name__}_{self.id}"

    def first_available_location(self) -> WarehouseLocation:
        """
        Cerchiamo la locazione in cui mettere la unit lod.
        Consideriamo solo le locazioni vuote e senza future unit load.
        """
        for location in self.locations:
            if location.is_empty and len(location.future_unit_loads) == 0:
                return location

    def first_available_location_for_warmup(self, unit_load):
        for location in self.locations:
            if location.is_empty:
                return location
            if location.is_half_full and location.product == unit_load.product:
                return location

    def create_input_operation(self, *, unit_load: T, location: WarehouseLocation, priority: int) -> InputOperation:
        input_operation = InputOperation(unit_load=unit_load, location=location, priority=priority)
        self._input_operations.append(input_operation)
        return input_operation

    @as_process
    def load_ant(self, *, feeding_operation: FeedingOperation):
        """
        Warehouse Output Process.

        Given a FeedingOperation, load the ant with the required unit load,
        once it is available from the Output Conveyor.
        """

        n_non_empty_locations = sum(not location.is_empty for location in self.locations)
        # n_non_empty_locations = sum(location.n_cases for location in self.locations)
        self._saturation_history.append((self.env.now, n_non_empty_locations / self.n_locations))

        yield self.env.timeout(self.load_time)

        output_operation = yield self.output_conveyor.get(
            lambda output_operation: output_operation.unit_load == feeding_operation.unit_load
        )
        if output_operation.unit_load != feeding_operation.unit_load:
            raise ValueError("Unit load mismatch")

        yield feeding_operation.ant.load(unit_load=feeding_operation.unit_load)

    @as_process
    def unload_ant(self, *, ant: Ant, input_operation: InputOperation):
        """
        Warehouse Input Process.

        Given an Ant and an InputOperation, unload the unit load from the ant and put it on the input conveyor,
        once it is available.
        """

        n_non_empty_locations = sum(not location.is_empty for location in self.locations)
        # n_non_empty_locations = sum(location.n_cases for location in self.locations)
        self._saturation_history.append((self.env.now, n_non_empty_locations / self.n_locations))

        with self.input_service_point.request(
            priority=input_operation.priority, preempt=False
        ) as input_service_point_request:
            self.queue_stats.append((self.env.now, len(self.input_service_point.queue)))
            yield input_service_point_request
            yield self.input_conveyor.put((input_operation,))
            yield self.env.timeout(self.load_time)
            yield ant.unload()
            ant.release_current()

    def get(self, *, feeding_operation: FeedingOperation) -> simulatte.typings.ProcessGenerator:
        """
        Warehouse Main internal retrieval process.

        Given a FeedingOperation, the specific implementation of the method must handle all the steps
        to retrieve the unit load from within the warehouse from a specific location, until the loading of the
        unit load on the Output Conveyor.
        """
        raise NotImplementedError

    def load(
        self, *, unit_load: T, location: WarehouseLocation, ant: Ant, priority: int
    ) -> simulatte.typings.ProcessGenerator:
        """
        Warehouse Main internal loading process.

        Given a UnitLoad, the specific implementation of the method must handle all the steps
        needed to load the unit load on the Input Conveyor, and then delegate the loading of the unit laod inside
        the warehouse to the ._put method.
        """
        raise NotImplementedError

    def _put(self, *args, **kwargs):
        raise NotImplementedError

    def replenishment_started(self, *, product: Product, process) -> None:
        self._replenishment_processes[product.id].append(process)

    def book_location(self, *, location: WarehouseLocation, unit_load: Tray | Pallet) -> None:
        location.freeze(unit_load=unit_load)



from __future__ import annotations

from typing import TYPE_CHECKING, Hashable, Any

if TYPE_CHECKING:
    from simpy.resources.store import Store

from simpy.resources.store import StorePut


class HashStorePut(StorePut):
    """
    Request to put an *item* into an *HashStore*.
    Must provide the key mapping the item to put.
    """

    def __init__(self, resource: Store, key: Hashable, item: Any) -> None:
        self.key = key
        super().__init__(resource, item)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Generic, Sequence, TypeVar

from simpy.resources.store import FilterStore, Store

import simulatte

if TYPE_CHECKING:
    from simpy import Environment


T = TypeVar("T")


class SequentialStore(Generic[T]):
    """
    The SequentialStore implements a FIFO queue (items are stored in the order in which they are put).

    The SequentialStore is implemented as a combination of a FilterStore and a Store.
    The FilterStore is used to store the next item to be yielded,
    while the Store is used to store the remaining items.

    The retrieved element must satisfy the filter function applied to the FilterStore.
    """

    def __init__(self, env: Environment, capacity: int = float("inf")) -> None:
        if capacity <= 1:
            raise ValueError("Capacity of SequentialStore must be grater than 1.")

        self.env = env
        self._capacity = capacity
        self._internal_store = Store(env, capacity=capacity - 1)
        self._output = FilterStore(env, capacity=1)

    @property
    def capacity(self) -> int:
        return self._capacity

    @property
    def internal_store_level(self) -> int:
        return len(self._internal_store.items)

    @property
    def output_level(self) -> int:
        return len(self._output.items)

    @property
    def level(self) -> int:
        return self.internal_store_level + self.output_level

    @property
    def items(self) -> Sequence[T]:
        return self._output.items + self._internal_store.items

    @simulatte.as_process
    def put(self, item: T):
        if self.output_level == 0:
            yield self._output.put(item)
        else:
            yield self._internal_store.put(item)

    @simulatte.as_process
    def get(self, filter_: Callable) -> T:
        # Get the item from the output position
        item = yield self._output.get(filter_)

        # Eventually move the next item in the internal store to the output position
        if self.internal_store_level > 0:
            next_item = yield self._internal_store.get()
            yield self._output.put(next_item)

        # return retrieved item
        return item



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .multi_store import MultiStore

from simpy.resources.base import Get


class MultiStoreGet(Get):
    """Request to get *n* items from a *MultiStore*."""

    def __init__(self, store: MultiStore, n: int = 1) -> None:
        self.n = n
        super().__init__(store)



from __future__ import annotations

from typing import Generic, TypeVar

from simpy.resources.store import Store as SimpyStore

T = TypeVar("T")


class Store(SimpyStore, Generic[T]):
    def put(self, item: T) -> None:
        ...

    def get(self) -> T:
        ...



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from simulatte.unitload import Pallet

    from .physical_position import PhysicalPosition
    from .warehouse_location import WarehouseLocation


class LocationBusy(Exception):
    def __init__(self, location: WarehouseLocation):
        self.location = location
        super().__init__(location)

    def __str__(self) -> str:
        return f"{self.location} is busy"


class LocationEmpty(Exception):
    def __init__(self, location: WarehouseLocation):
        self.location = location
        super().__init__(f"Location {location} is empty")


class PhysicalPositionBusy(Exception):
    def __init__(self, physical_position: PhysicalPosition):
        self.physical_position = physical_position
        super().__init__(f"Physical position {physical_position} is busy")


class PhysicalPositionEmpty(Exception):
    def __init__(self, physical_position: PhysicalPosition):
        self.physical_position = physical_position
        super().__init__(f"Physical position {physical_position} is empty")


class IncompatibleUnitLoad(Exception):
    def __init__(self, unit_load: Pallet, location: WarehouseLocation):
        self.unit_load = unit_load
        self.location = location
        super().__init__(f"Unit load {unit_load} is not compatible with location {location}")



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from simulatte.location import Location
    from simulatte.system import System


class Distance:
    def __init__(self, system: System, from_: Location, to: Location) -> None:
        self.system = system
        self.from_ = from_
        self.to = to

    @property
    def as_distance(self) -> float:
        raise NotImplementedError

    @property
    def as_time(self) -> float:
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar, Generic

import matplotlib.pyplot as plt
import simpy

import simulatte

if TYPE_CHECKING:
    from simulatte.stores import WarehouseLocation, WarehouseStore

T = TypeVar("T", bound=simulatte.unitload.CaseContainer)


class Traslo(simpy.PriorityResource, Generic[T]):
    def __init__(
        self, *, store: WarehouseStore, x: int, y: int, speed_x: float, speed_y: float, load_time: float
    ) -> None:
        self.env = simulatte.Environment()
        super().__init__(self.env, capacity=1)

        self.store = store
        self.x = x
        self.y = y
        self.speed_x = speed_x
        self.speed_y = speed_y
        self.load_time = load_time

        self._handling_time = 0
        self._saturation_history: simulatte.typings.History[float] = [(0, 0)]
        self._unit_load: T | None = None

    @property
    def handling_time(self) -> float:
        return self._handling_time

    @handling_time.setter
    def handling_time(self, handling_time: float) -> None:
        self._handling_time = handling_time
        self._saturation_history.append((self.env.now, self.saturation))

    @property
    def saturation(self) -> float:
        return self._handling_time / self.env.now

    @property
    def unit_load(self) -> T | None:
        return self._unit_load

    @unit_load.setter
    def unit_load(self, unit_load: T | None) -> None:
        if self._unit_load is not None and unit_load is not None:
            raise RuntimeError("The traslo is already loaded.")

        if self._unit_load is None and unit_load is None:
            raise RuntimeError("The traslo is not loaded.")

        self._unit_load = unit_load

    @simulatte.as_process
    def move(self, *, location: WarehouseLocation) -> None:
        time_x = abs(self.x - location.x) * location.width / self.speed_x
        time_y = abs(self.y - location.y) * location.height / self.speed_y
        t = max(time_x, time_y)
        yield self.env.timeout(max(time_x, time_y))
        self.handling_time += t
        self.x, self.y = location.x, location.y

    @simulatte.as_process
    def load(self, *, unit_load: T) -> None:
        self.unit_load = unit_load
        yield self.env.timeout(self.load_time)
        self.handling_time += self.load_time

    @simulatte.as_process
    def unload(self) -> None:
        self.unit_load = None
        yield self.env.timeout(self.load_time)
        self.handling_time += self.load_time

    def plot(self) -> None:
        x = [t / 3600 for t, _ in self._saturation_history]
        y = [s * 100 for _, s in self._saturation_history]
        plt.plot(x, y)
        plt.xlabel("Time [h]")
        plt.ylabel(f"Saturation [%]")
        plt.title(f"{self.store.name} {self.__class__.__name__} Productivity")
        plt.ylim([0, 100])
        plt.show()



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from .filter_multi_store import FilterMultiStore

from simpy.resources.base import Get


class FilterMultiStoreGet(Get):
    """
    Request to get items from a FilterMultiStore based on a 'filter' callable.
    """

    def __init__(self, store: FilterMultiStore, filter: Callable):
        self.filter = filter
        super().__init__(store)



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence, Callable, Any

if TYPE_CHECKING:
    from simpy import Environment

from ..sequential_store import SequentialStore
from ..multi_store import MultiStore


class SequentialMultiStore(SequentialStore):
    """
    An instance of this class represents a SequentialStore where is
    possible to put more items at the same time (as in a MultiStore).

    It is not possible to get more elements at the same time because a
    one-piece-flow sequence must be respected.
    """

    def __init__(self, env: Environment, capacity: int = float("inf")):
        super().__init__(env, capacity)
        # Overwrite the store with a MultiStore instance
        self._internal_store = MultiStore(env, capacity - 1)

    def _do_put(self, items: Sequence):
        if len(items) >= self.capacity:
            raise Exception(f"Items to store exceed the capacity {self.capacity}.")

        if self.output_level == 0:
            yield self._output.put(items[0])
            if len(items) > 1:
                yield self._internal_store.put(items[1:])
        else:
            yield self._internal_store.put(items)

    def _do_get(self, filter_: Callable):
        # Get the item from the exit position
        item = yield self._output.get(filter_)

        # Eventually move the next item in the internal store to the output position
        if self.internal_store_level > 0:
            next_item = yield self._internal_store.get(1)
            self._output.put(next_item[0])

        # return retrieved item
        return item



from __future__ import annotations

from typing import TYPE_CHECKING

from simulatte.logger.logger import EventPayload
from ..observables import Observable
from .base_area import Area, T

if TYPE_CHECKING:
    from simulatte.picking_cell import PickingCell


class ObservableArea(Area[T], Observable):
    def __init__(self, *, cell: PickingCell, capacity: int = float("inf")):
        Area.__init__(self, cell=cell, capacity=capacity)
        Observable.__init__(self, system=cell.system)

    def remove(self, item: T) -> None:
        Area.remove(self, item)
        payload = EventPayload(event=f"ACTIVATING {self.__class__.__name__} SIGNAL", type=1)
        self.trigger_signal_event(payload=payload)



import heapq


class PriorityItem:
    """An object with the priority attribute"""

    def __init__(self, key):
        self.key = key

    def __lt__(self, other):
        return self.key < other.key

    @property
    def priority(self):
        return self.key


class PriorityQueue:
    """
    An instance of this class represents a priority queue for items
    that already have the attribute prority.
    """

    def __init__(self, items=None, maxlen=float("inf")):
        self.items = items
        self.maxlen = maxlen
        if items is not None:
            self.items = [(i.priority, i) for i in items]
        heapq.heapify(self.items)

    def __len__(self):
        return len(self.items)

    def push(self, item):
        if self.items.__len__() >= self.maxlen:
            raise Exception("Capacity of PriorityQueue exceeded.")
        heap_item = (item.priority, item)
        heapq.heappush(self.items, heap_item)

    def pop(self):
        item = heapq.heappop(self.items)
        return item[1]

    def pushpop(self, item):
        heap_item = (item.priority, item)
        to_return = heapq.heappushpop(heap_item)
        return to_return[1]



import math

from .warehouse_location import WarehouseLocation


def euclidean(location_a: WarehouseLocation, location_b: WarehouseLocation) -> float:
    """
    The euclidean distance between two WarehouseLocations.
    """

    loc_a_x, loc_a_y = location_a.coordinates
    loc_b_x, loc_b_y = location_b.coordinates
    return math.sqrt((loc_a_x - loc_b_x) ** 2 + ((loc_a_y - loc_b_y) ** 2))


def manhattan(location_a: WarehouseLocation, location_b: WarehouseLocation) -> float:
    """
    The manhattan distance between two WarehouseLocations.
    """

    loc_a_x, loc_a_y = location_a.coordinates
    loc_b_x, loc_b_y = location_b.coordinates
    return abs(loc_a_x - loc_a_y) + abs(loc_b_x - loc_b_y)



from __future__ import annotations

from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:
    from simulatte import System
    from simulatte.picking_cell import PickingCell


class CellsManager:
    def __init__(self, system: System, *cells: PickingCell) -> None:
        self.system = system
        self._cells = list(cells)

    def __call__(self, cell: PickingCell) -> CellsManager:
        self._cells.append(cell)
        return self

    @property
    def cells(self):
        return self._cells

    def get_best_picking_cell(self, *, cls: Type[PickingCell] | None = None) -> PickingCell:
        cells = self.cells
        if cls is not None:
            cells = (c for c in cells if isinstance(c, cls))

        cell = min(cells, key=lambda c: len(c.feeding_area))
        if cell is None:
            raise ValueError("No picking cell available")
        return cell



from simpy.core import BoundClass
from simpy.resources.resource import Release, Request, Resource


class OccupationRequest(Request):
    """
    An instance of this request is a request made to a position
    with an associated FeedingOperation.

    It is instantiated when an ant wants to occupy a Position.
    """

    def __init__(self, resource, operation):
        """
        Initialize.

        :param ant: The ant that wants to occupy the resource
        :param priority: The priority of the request
        """
        self.operation = operation
        super(OccupationRequest, self).__init__(resource)


class Position(Resource):
    """
    An instance of this class represents a position that can be booked and occupied by an ant inside the picking cell.
    """

    # Method to require the position
    request = BoundClass(OccupationRequest)

    # Method to release the position
    release = BoundClass(Release)

    def __init__(self, name: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name

    def __repr__(self) -> str:
        return self.name

    @property
    def busy(self):
        return len(self.users) > 0

    @property
    def empty(self):
        return len(self.users) == 0

    def release_current(self):
        if len(self.users) == 0:
            raise Exception("Position cannot release unexisting request.")
        self.release(self.users[0])

    @property
    def id(self):
        if self.operation:
            return self.operation.layers[0].id

    @property
    def operation(self):
        if len(self.users) == 0:
            return None
        return self.users[0].operation

    @property
    def product(self):
        if self.operation:
            return self.operation.product

    @property
    def ant(self):
        if len(self.users) == 0:
            return None
        return self.users[0].operation.ant



from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING

from .case_container import CaseContainer
from .tray import Tray
from .wood_board import WoodBoard

if TYPE_CHECKING:
    from simulatte.products import Product


class Pallet(CaseContainer):
    def __init__(self, *layers: Tray, wood_board=False) -> None:
        if wood_board:
            self.layers = deque((WoodBoard(), *layers))
        else:
            self.layers = deque(layers)
        self.location = None

    @classmethod
    def by_product(cls, *, product: Product) -> Pallet:
        return cls(
            *(Tray(product=product, n_cases=product.cases_per_layer) for _ in range(product.layers_per_pallet)),
        )

    def __repr__(self) -> str:
        return f"Pallet(*{self.layers})"

    @classmethod
    def single_tray(cls, *, product: Product, n_cases: int | None = None) -> Pallet:
        return Pallet(*[Tray(product=product, n_cases=n_cases if n_cases is not None else product.cases_per_layer)])

    @property
    def upper_layer(self) -> Tray | None:
        if self.n_layers > 0:
            return self.layers[-1]

    @property
    def product(self) -> Product | None:
        """The product accessible on the unit load --i.e., the product on the last layer."""
        if self.upper_layer is not None:
            return self.upper_layer.product

    @property
    def n_cases(self) -> int:
        """The number of cases on the unit load"""
        trays = (layer for layer in self.layers if isinstance(layer, Tray))
        return sum(tray.n_cases for tray in trays)

    @property
    def n_layers(self) -> int:
        """The number of Trays on the unit load"""
        return sum(isinstance(layer, Tray) for layer in self.layers)

    def remove_case(self) -> None:
        """
        Removes a single case from the most accessible layer
        """
        self.upper_layer.n_cases -= 1

    def remove_layer(self) -> None:
        """
        Removes the most accessible layer from the unit load.
        """
        self.layers.pop()

    def add_layer(self, *, product: Product, n_cases: int) -> None:
        """
        Adds a layer to the unit load.
        """
        self.layers.append(Tray(product=product, n_cases=n_cases))



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from simulatte.products import Product
from simulatte.unitload import Pallet
from simulatte.utils import Identifiable

if TYPE_CHECKING:
    from simulatte.picking_cell import FeedingOperation


class Request:
    product: Product
    n_cases: int

    def __init__(self) -> None:
        self.feeding_operations: list[FeedingOperation] = []
        self.picked_n_cases = 0

    @property
    def remaining_to_pick_n_cases(self) -> int:
        return self.n_cases - self.picked_n_cases


class CaseRequest(Request):
    def __init__(self, product: Product) -> None:
        super().__init__()
        self.product = product

    def __repr__(self) -> str:
        return f"CaseRequest(product={self.product})"

    @property
    def n_cases(self) -> int:
        return 1


class ProductRequest(Request):
    def __init__(self, product: Product, n_cases: int) -> None:
        if n_cases > product.cases_per_layer:
            raise ValueError("A ProductRequest cannot exceed the product cases per layer")

        super().__init__()
        self.product = product
        self.sub_requests = [CaseRequest(product=product) for _ in range(n_cases)]
        self.processed = False
        self.pallet_request: PalletRequest | None = None

    def __repr__(self) -> str:
        return f"ProductRequest(product={self.product}, n_cases={self.n_cases})"

    @property
    def n_cases(self) -> int:
        return len(self.sub_requests)


class LayerRequest(Request):
    def __init__(self, *product_requests: ProductRequest) -> None:
        super().__init__()

        n_cases_requested = sum(r.n_cases for r in product_requests)

        if n_cases_requested > sum(r.product.cases_per_layer for r in product_requests):
            raise ValueError("Overflow of cases in the LayerRequest")

        self.sub_requests = list(product_requests)
        self.pallet_request: PalletRequest | None = None

    def __repr__(self) -> str:
        return f"LayerRequest(sub_requests={self.sub_requests})"

    @property
    def processed(self) -> bool:
        return all(product_request.processed for product_request in self.sub_requests)

    @property
    def n_cases(self) -> int:
        return sum(product_request.n_cases for product_request in self.sub_requests)

    @property
    def has_single_product_request(self) -> bool:
        """
        Returns True if the LayerRequest is composed of a single ProductRequest.
        (the LayerRequest should be processed by a LayerPickingCell)
        """
        return len(self.sub_requests) == 1

    @property
    def product(self) -> Product:
        if not self.has_single_product_request:
            raise ValueError("The LayerRequest contains more than one ProductRequest")
        return self.sub_requests[0].product

    @property
    def products(self) -> Iterable[Product]:
        for product_request in self.sub_requests:
            yield product_request.product

    def add(self, *, product_request: ProductRequest) -> LayerRequest:
        self.sub_requests.append(product_request)
        return self


class PalletRequest(Request, metaclass=Identifiable):
    def __init__(self, *layer_requests: LayerRequest, wood_board=False) -> None:
        super().__init__()
        self.sub_requests = list(layer_requests)
        self.unit_load = Pallet(wood_board=wood_board)

        for layer_request in self.sub_requests:
            layer_request.pallet_request = self
            for product_request in layer_request.sub_requests:
                product_request.pallet_request = self

        self._start_time = None
        self._end_time = None

    def __repr__(self):
        return f"PalletRequest(id={self.id})"

    def __iter__(self) -> Iterable[LayerRequest]:
        return (layer_request for layer_request in self.sub_requests if not layer_request.processed)

    @property
    def n_cases(self) -> int:
        return sum(layer_request.n_cases for layer_request in self.sub_requests)

    @property
    def products(self) -> Iterable[Product]:
        for layer_request in self.sub_requests:
            yield from layer_request.products

    @property
    def is_for_layer_picking_cell(self) -> bool:
        """
        Returns True if the PalletRequest has to be processed *ONLY* by a LayerPickingCell.
        (all LayerRequests are composed of **one** ProductRequest)
        """
        return all(layer_request.has_single_product_request for layer_request in self.sub_requests)

    @property
    def is_for_case_picking_cell(self) -> bool:
        """
        Returns True if the PalletRequest has to be processed *ONLY* by a CasePickingCell.
        (all LayerRequests are composed of **more than one** ProductRequest)
        """
        return all(not layer_request.has_single_product_request for layer_request in self.sub_requests)

    def assigned(self, time: int) -> None:
        self._start_time = time

    def completed(self, time: int) -> None:
        self._end_time = time

    @property
    def lead_time(self) -> float | None:
        if self._start_time is not None and self._end_time is not None:
            return self._end_time - self._start_time



from __future__ import annotations

import collections
from typing import TYPE_CHECKING, Iterable

from IPython.display import Markdown, display
from simpy import Process
from tabulate import tabulate

from simulatte.location import Location
from simulatte.logger.logger import EventPayload
from simulatte.operations import FeedingOperation
from simulatte.picking_cell.areas import FeedingArea, InternalArea, StagingArea
from simulatte.picking_cell.observers import FeedingObserver, StagingObserver
from simulatte.picking_cell.observers.internal_observer import InternalObserver
from simulatte.requests import PalletRequest, ProductRequest
from simulatte.resources import MonitoredResource
from simulatte.simpy_extension import SequentialStore
from simulatte.utils.utils import as_process

if TYPE_CHECKING:
    from simulatte.requests import Request
    from simulatte.resources.store import Store
    from simulatte.system import System
    from simulatte.typings import ProcessGenerator


class PickingCell:
    def __init__(
        self,
        *,
        system: System,
        input_queue: Store[PalletRequest],
        output_queue: SequentialStore[PalletRequest],
        building_point: MonitoredResource,
        feeding_area_capacity: int,
        staging_area_capacity: int,
        internal_area_capacity: int,
        register_main_process: bool = True,
    ):
        self.system = system
        self.system.cells_manager(self)

        self.input_location = Location(name=f"{self.__class__.__name__} Input")
        self.input_queue = input_queue
        self.output_location = Location(name=f"{self.__class__.__name__} Output")
        self.output_queue = output_queue
        self.building_point = building_point

        self.feeding_operations: list[FeedingOperation] = []

        self.feeding_area = FeedingArea(cell=self, capacity=feeding_area_capacity)
        self.feeding_observer = FeedingObserver(system=self.system, observable_area=self.feeding_area)

        self.staging_area = StagingArea(cell=self, capacity=staging_area_capacity)
        self.staging_observer = StagingObserver(system=self.system, observable_area=self.staging_area)

        self.internal_area = InternalArea(cell=self, capacity=internal_area_capacity)
        self.internal_observer = InternalObserver(system=self.system, observable_area=self.internal_area)

        self.picking_requests_queue: collections.deque[Request] = collections.deque()

        self.feeding_operation_map: dict[Request, list[FeedingOperation]] = collections.defaultdict(list)

        self.current_pallet_request: PalletRequest | None = None

        self.staging_location = Location(name=f"{self.__class__.__name__} StagingAreaLocation")
        self.internal_location = Location(name=f"{self.__class__.__name__} InternalAreaLocation")

        self._productivity_history: list[tuple[float, float]] = []

        self.pallet_requests_assigned: list[PalletRequest] = []
        self.pallet_requests_done: list[PalletRequest] = []

        self._main: Process | None = None
        if register_main_process:
            self._main = self.main()

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def productivity(self) -> float:
        """
        Return the productivity of the PickingCell, expressed as number of PalletRequest completed per unit of time.
        """
        return len(self.pallet_requests_done) / self.system.env.now

    def register_feeding_operation(self, *, feeding_operation: FeedingOperation) -> None:
        """
        Register a FeedingOperation into the PickingCell.

        Updates the PickingCell.feeding_operation_map,
        which maps which FeedingOperation is associated to which ProductRequest.
        It also adds the FeedingOperation to the FeedingArea.
        """
        self.feeding_operation_map[feeding_operation.picking_request].append(feeding_operation)
        self.feeding_area.append(feeding_operation, exceed=True)

    @as_process
    def _retrieve_feeding_operation(self, picking_request: Request) -> FeedingOperation:
        """
        FIXME: PORCATA COLOSSALE!!!
        Tocca usare un processo con attesa per evitare che ci sia una race-condition tra il momento in cui
        si registra l'associazione tra feeding_operation e picking_request e il momento in cui si
        interroga l'associazione per recuperare la feeding_operation.
        """
        while picking_request not in self.feeding_operation_map:
            yield self.system.env.timeout(0.1)
        return self.feeding_operation_map[picking_request]

    @as_process
    def _process_product_request(
        self, *, product_request: ProductRequest, pallet_request: PalletRequest
    ) -> ProcessGenerator:
        """
        Main process which manages the handling of a ProductRequest.

        To be concretely implemented in a subclass.
        """
        raise NotImplementedError

    @as_process
    def let_ant_in(self, *, feeding_operation: FeedingOperation) -> ProcessGenerator:
        """
        Manage the internal physical movements of an Ant which moves from the StagingArea to the InternalArea.
        """

        # If necessary, wait for the assigned InternalArea PreUnloadPosition to be free
        pre_unload_position_request = None
        if feeding_operation.pre_unload_position is not None:
            pre_unload_position_request = feeding_operation.pre_unload_position.request(operation=feeding_operation)
            yield pre_unload_position_request

        # Wait for the assigned InternalArea UnloadPosition to be free
        unload_position_request = feeding_operation.unload_position.request(operation=feeding_operation)
        yield unload_position_request

        if pre_unload_position_request is not None:
            # Release the assigned InternalArea PreUnloadPosition
            feeding_operation.pre_unload_position.release(pre_unload_position_request)

        # Move the Ant from the StagingArea to the InternalArea
        yield feeding_operation.ant.move_to(system=self.system, location=self.internal_location)

        # Housekeeping
        feeding_operation.ready_for_unload()
        feeding_operation.ant.waiting_to_be_unloaded()

    @as_process
    def let_ant_out(self, *, feeding_operation: FeedingOperation) -> ProcessGenerator:
        feeding_operation.ant.picking_ends()

        # free the UnloadPosition associated to the FeedingOperation
        feeding_operation.unload_position.release_current()

        self.internal_area.remove(feeding_operation)

        if feeding_operation.unit_load.n_cases > 0:
            store = feeding_operation.store
            ant = feeding_operation.ant
            yield ant.move_to(system=self.system, location=store.input_location)
            yield self.system.stores_manager.load(store=store, ant=ant)
        else:
            from simulatte.ant import ant_rest_location

            # otherwise, move the Ant to the rest location
            yield feeding_operation.ant.move_to(system=self.system, location=ant_rest_location)
            yield feeding_operation.ant.unload()
            feeding_operation.ant.release_current()

        feeding_operation.ant.idle()
        feeding_operation.ant.mission_ended()

    @as_process
    def put(self, *, pallet_request: PalletRequest) -> ProcessGenerator:
        """
        Assigns a PalletRequest to the PickingCell.
        The PalletRequest is stored within the input queue.
        Moreover, the PalletRequest is then decomposed in picking requests,
        which are placed in the picking requests queue.
        Eventually, the FeedingArea signal event is triggered.
        """
        self.pallet_requests_assigned.append(pallet_request)
        yield self.input_queue.put(pallet_request)
        self.picking_requests_queue.extend(self.iter_pallet_request(pallet_request=pallet_request))

        # Triggers the feeding area signal event
        # until the feeding area is full or there are no more picking requests to be processed
        while not self.feeding_area.is_full and len(self.picking_requests_queue) > 0:
            payload = EventPayload(event="ACTIVATING FEEDING AREA SIGNAL", type=0)
            self.feeding_area.trigger_signal_event(payload=payload)
            yield self.system.env.timeout(0.001)

    @as_process
    def get(self, pallet_request: PalletRequest) -> ProcessGenerator[PalletRequest]:
        """
        Retrieves a completed PalletRequest.
        """
        pallet_request = yield self.output_queue.get(lambda e: e == pallet_request)
        return pallet_request

    @staticmethod
    def iter_pallet_request(*, pallet_request: PalletRequest) -> Iterable[ProductRequest]:
        """
        Iterates over the  ProductRequest of a PalletRequest.
        """
        for layer_request in pallet_request.sub_requests:
            for product_request in layer_request.sub_requests:
                yield product_request

    @as_process
    def main(self) -> ProcessGenerator:
        """
        Main PickingCell process.

        Waits for a PalletRequest to be handled.
        Once a PalletRequest is obtained, it waits for the availability of the BuildingPoint.
        Generates a handling process for each ProductRequest to be handled within the PalletRequest.
        Waits for the processing of all ProductRequest.
        Once finished, positions the completed PalletRequest in the output queue of the cell,
        and asks the System to handle the retrieval of the finished PalletRequest.
        """
        while True:
            # Wait for a PalletRequest to be handled
            pallet_request: PalletRequest = yield self.input_queue.get()
            pallet_request.assigned(time=self.system.env.now)
            self.current_pallet_request = pallet_request

            with self.building_point.request() as building_point_request:
                # Wait for the availability of the BuildingPoint
                yield building_point_request
                self.building_point.current_pallet_request = pallet_request

                for product_request in self.iter_pallet_request(pallet_request=pallet_request):
                    yield self._process_product_request(product_request=product_request, pallet_request=pallet_request)

                # Once finished, positions the completed PalletRequest in the output queue of the cell
                yield self.output_queue.put(pallet_request)

                # Housekeeping
                self.pallet_requests_done.append(pallet_request)
                pallet_request.completed(time=self.system.env.now)
                self._productivity_history.append((self.system.env.now, self.productivity))

                # Ask the System to handle the retrieval of the finished PalletRequest
                self.system.retrieve_from_cell(cell=self, pallet_request=pallet_request)

    def summary(self):
        display(Markdown("## Performance Summary"))

        headers = ["KPI", "Valore", "U.M."]
        table = [
            ["Ore simulate", f"{self.system.env.now / 60 / 60:.2f}", "h"],
            [
                "Produttività Cella",
                f"{self.productivity * 60 * 60:.2f}",
                "[PalletRequest/h]",
            ],
            [
                "Produttività Cella",
                f"{self.productivity * 60 * 60 * 4 * 10:.2f}",
                "[Cases/h]",
            ],
            [
                "Produttività Robot",
                f"{self.robot.productivity * 60 * 60:.2f}",
                "[Cases/h]",
            ],
        ]
        print(tabulate(table, headers=headers, tablefmt="fancy_grid"))
        display(Markdown("## Robot"))
        self.robot.plot()
        display(Markdown("## Aree logiche/fisiche"))
        self.feeding_area.plot()
        self.staging_area.plot()
        self.internal_area.plot()



from __future__ import annotations

from typing import TYPE_CHECKING

import simulatte

if TYPE_CHECKING:
    from simulatte.stores.warehouse_location import WarehouseLocation
    from simulatte.unitload import Pallet


class Operation:
    def __init__(self, *, unit_load: Pallet, location: WarehouseLocation, priority: int) -> None:
        self.env = simulatte.Environment()
        self.unit_load = unit_load
        self.location = location
        self.priority = priority

    @property
    def position(self) -> int:
        return self.location.x

    @property
    def floor(self) -> int:
        return self.location.y

    def __eq__(self, other) -> bool:
        return self.unit_load == other.unit_load


class InputOperation(Operation):
    """Warehouse input operation"""

    def __init__(self, *, unit_load: Pallet, location: WarehouseLocation, priority: int) -> None:
        super().__init__(unit_load=unit_load, location=location, priority=priority)
        self.lift_process = None
        self.lifted = self.env.event()


class OutputOperation(Operation):
    """Warehouse output operation"""

    pass



from .ants_policy import AntSelectionPolicy, MinorWorkloadAntsPolicy
from .location_policy import LocationPolicy, ClosestLocationPolicy
from .unit_load_policy import UnitLoadPolicy, ClosestUnitLoadPolicy, MultiStoreLocationPolicy



from .exceptions import (
    IncompatibleUnitLoad,
    LocationBusy,
    LocationEmpty,
    PhysicalPositionBusy,
    PhysicalPositionEmpty,
)
from .physical_position import PhysicalPosition
from .warehouse_location import WarehouseLocation



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence, Any

if TYPE_CHECKING:
    from .multi_store import MultiStore

from simpy.resources.base import Put


class MultiStorePut(Put):
    """Request to put *items* into a *MultiStore*."""

    def __init__(self, store: MultiStore, items: Sequence[Any]) -> None:
        self.items = items
        super().__init__(store)



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

if TYPE_CHECKING:
    from simulatte.ant import Ant


class AntSelectionPolicy:
    def __call__(self, *, ants: Sequence[Ant], exceptions: Sequence[Ant] | None = None) -> Ant:
        raise NotImplementedError


class MinorWorkloadAntsPolicy(AntSelectionPolicy):
    @staticmethod
    def sorter(ant: Ant) -> tuple[int, int]:
        return len(ant.users), len(ant.queue)

    def __call__(self, *, ants: Sequence[Ant], exceptions: Sequence[Ant] | None = None) -> Ant | None:
        exceptions = exceptions or set()
        return min((ant for ant in ants if ant not in exceptions), key=self.sorter, default=None)



from simpy.core import BoundClass
from simpy.resources.store import Store

from .hash_store_get import HashStoreGet
from .hash_store_put import HashStorePut


class HashStore(Store):
    """
    Like a FilterStore but the research is more efficient
    because based on a HashMap.

    The HashStore works like a 'simpy.resource.store.FilterStore',
    but allows the storage and retrieval of an item based on efficient mapping.
    """

    get = BoundClass(HashStoreGet)

    put = BoundClass(HashStorePut)

    def __init__(self, env, capacity=float("inf")) -> None:
        if capacity <= 0:
            raise ValueError("capacity must be > 0.")

        super().__init__(env, capacity)
        self.items = {}

    def _do_put(self, event: HashStorePut) -> None:
        """
        Put an item into the store, if there is enough space.
        The storage of the item is based on the key passed in the 'event.key' attribute.
        """
        if self.level < self._capacity:
            self.items[event.key] = event.item
            event.succeed()
        return None

    def _do_get(self, event: HashStoreGet) -> None:
        """
        Get an item from the store.
        The retrieval of the item is based on the key passed in the 'event.key' attribute.
        If the key is not found, the event is not triggered.
        Optionally, a KeyError can be raised if the key is not found.
        """
        try:
            item = self.items.pop(event.key)
            event.succeed(item)
        except KeyError as e:
            if event.raise_missing:
                raise e
            else:
                return None
        return None

    @property
    def level(self) -> int:
        """
        The number of items currently in the store.
        """
        return len(self.items)



from __future__ import annotations

from typing import TYPE_CHECKING, Hashable

if TYPE_CHECKING:
    from simpy.resources.store import Store

from simpy.resources.store import StoreGet


class HashStoreGet(StoreGet):
    """
    Request to get an *item* from an *HashStore*.
    Must provide the key of the item to get.
    Optionally, a KeyError can be raised if the key is not found within the *HashStore*.
    """

    def __init__(self, resource: Store, key: Hashable, raise_missing=False) -> None:
        self.key = key
        self.raise_missing = raise_missing
        super().__init__(resource)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['simulatte',
 'simulatte.ant',
 'simulatte.buffer',
 'simulatte.demand',
 'simulatte.demand.generators',
 'simulatte.distance',
 'simulatte.events',
 'simulatte.exceptions',
 'simulatte.logger',
 'simulatte.operations',
 'simulatte.picking_cell',
 'simulatte.picking_cell.areas',
 'simulatte.picking_cell.observables',
 'simulatte.picking_cell.observers',
 'simulatte.resources',
 'simulatte.service_point',
 'simulatte.simpy_extension',
 'simulatte.simpy_extension.filter_multi_store',
 'simulatte.simpy_extension.hash_store',
 'simulatte.simpy_extension.multi_store',
 'simulatte.simpy_extension.sequential_multi_store',
 'simulatte.simpy_extension.sequential_store',
 'simulatte.stores',
 'simulatte.stores.warehouse_location',
 'simulatte.system',
 'simulatte.system.managers',
 'simulatte.system.policies',
 'simulatte.timings',
 'simulatte.typings',
 'simulatte.unitload',
 'simulatte.utils']

package_data = \
{'': ['*']}

install_requires = \
['ipython>=8.11.0,<9.0.0',
 'matplotlib>=3.6.2,<4.0.0',
 'pandas>=1.5.1,<2.0.0',
 'simpy>=4.0.1,<5.0.0',
 'tabulate>=0.9.0,<0.10.0']

setup_kwargs = {
    'name': 'simulatte',
    'version': '0.1.3',
    'description': '',
    'long_description': '![Coverage](https://raw.githubusercontent.com/dmezzogori/simulatte/main/coverage.svg)\n',
    'author': 'Davide Mezzogori',
    'author_email': 'dmezzogori@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from typing import TypeVar, Generator

from simpy import Event

ProcessReturn = TypeVar("ProcessReturn")
ProcessGenerator = Generator[Event, None, ProcessReturn | None]

HistoryValue = TypeVar("HistoryValue")
History = list[tuple[float, HistoryValue]]



from __future__ import annotations

import csv
from statistics import mean
from typing import TYPE_CHECKING, Sequence

from IPython.display import Markdown, display
from tabulate import tabulate

from simulatte.system.policies import AntSelectionPolicy

if TYPE_CHECKING:
    from simulatte.ant import Ant


class AntsManager:
    def __init__(self, ants: Sequence[Ant], ant_selection_policy: AntSelectionPolicy) -> None:
        self.ants = ants
        self._ant_selection_policy = ant_selection_policy

    @property
    def feeding_ants(self) -> tuple[Ant]:
        return tuple(ant for ant in self.ants if ant.kind == "feeding")

    @property
    def replenishment_ants(self) -> tuple[Ant]:
        return tuple(ant for ant in self.ants if ant.kind == "replenishment")

    def get_best_ant(self, exceptions: Sequence[Ant] | None = None) -> Ant:
        return self._ant_selection_policy(ants=self.ants, exceptions=exceptions)

    def export_mission_logs_csv(self, path: str) -> None:
        with open(path, "w") as csvfile:
            fieldnames = [
                "ant_id",
                "start_timestamp",
                "start_location",
                "end_timestamp",
                "end_location",
            ]
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()

            for ant in self.ants:
                for mission in ant.mission_logs:
                    writer.writerow(
                        {
                            "ant_id": ant.id,
                            "start_timestamp": mission.start_time,
                            "start_location": mission.start_location.name,
                            "end_timestamp": mission.end_time,
                            "end_location": mission.end_location.name,
                        }
                    )

    def summary(self) -> None:
        display(Markdown("## Ant fleet"))

        headers = [
            "ANT",
            "SAT\n[%]",
            "TOT WAIT\n[min]",
            "WAIT@AVSRS\nAVG [min]",
            "WAIT@FEED\nAVG [min]",
            "WAIT@STAG\nAVG [min]",
            "WAIT@INT\nAVG [min]",
            "WAIT@PICKING\nAVG [min]",
        ]
        table = []
        for ant in self.ants:
            saturation = f"{ant.saturation * 100:.2f}"
            total_waiting_time = f"{ant.waiting_time / 60:.2f}"
            waiting_at_avsrs = f"{(mean(ant.loading_waiting_times) if ant.loading_waiting_times else 0) / 60:.2f}"
            waiting_in_feeding_area = (
                f"{(mean(ant.feeding_area_waiting_times) if ant.feeding_area_waiting_times else 0) / 60:.2f}"
            )
            waiting_in_staging_area = (
                f"{(mean(ant.staging_area_waiting_times) if ant.staging_area_waiting_times else 0) / 60:.2f}"
            )
            waiting_in_internal_area = (
                f"{(mean(ant.unloading_waiting_times) if ant.unloading_waiting_times else 0) / 60:.2f}"
            )
            waiting_at_picking = f"{(mean(ant.picking_waiting_times) if ant.picking_waiting_times else 0) / 60:.2f}"

            table.append(
                [
                    ant.id,
                    saturation,
                    total_waiting_time,
                    waiting_at_avsrs,
                    waiting_in_feeding_area,
                    waiting_in_staging_area,
                    waiting_in_internal_area,
                    waiting_at_picking,
                ]
            )
        print(tabulate(table, headers=headers, tablefmt="fancy_grid"))

        headers = [
            "ANT",
            "SAT\n[%]",
            "TOT WAIT\n[min]",
            "WAIT@AVSRS\nTOT [min]",
            "WAIT@FEED\nTOT [min]",
            "WAIT@STAG\nTOT [min]",
            "WAIT@INT\nTOT [min]",
            "WAIT@PICKING\nTOT [min]",
            "SUM\n[min]",
        ]
        table = []
        for ant in self.ants:
            saturation = f"{ant.saturation * 100:.2f}%"
            total_waiting_time = f"{ant.waiting_time / 60:.2f}"
            waiting_at_avsrs = f"{sum(ant.loading_waiting_times) / 60:.2f}"
            waiting_in_feeding_area = f"{sum(ant.feeding_area_waiting_times) / 60:.2f}"
            waiting_in_staging_area = f"{sum(ant.staging_area_waiting_times) / 60:.2f}"
            waiting_in_internal_area = f"{sum(ant.unloading_waiting_times) / 60:.2f}"
            waiting_at_picking = f"{sum(ant.picking_waiting_times) / 60:.2f}"

            tot = (
                sum(ant.loading_waiting_times)
                + sum(ant.feeding_area_waiting_times)
                + sum(ant.staging_area_waiting_times)
                + sum(ant.unloading_waiting_times)
                + sum(ant.picking_waiting_times)
            )
            tot = f"{tot / 60:.2f}"

            table.append(
                [
                    ant.id,
                    saturation,
                    total_waiting_time,
                    waiting_at_avsrs,
                    waiting_in_feeding_area,
                    waiting_in_staging_area,
                    waiting_in_internal_area,
                    waiting_at_picking,
                    tot,
                ]
            )
        print(tabulate(table, headers=headers, tablefmt="fancy_grid"))



from __future__ import annotations

from .observable_area import ObservableArea
from ...operations import FeedingOperation


class FeedingArea(ObservableArea[FeedingOperation]):
    """
    Represent the logical area of feeding ants currently serving a picking cell.
    """

    pass



import simpy

import simulatte
from simulatte.location import Location


class ServicePoint(simpy.PriorityResource):
    """
    An instance of this class represents a ServicePoint: a position
    where ants go to be served.
    """

    def __init__(self, *, location: Location, capacity=1):
        """
        Initialise.

        :param env: The simulation environment
        :param loc: The node where the service point is placed.
        """
        self.env = simulatte.Environment()
        self.location = location
        super().__init__(self.env, capacity=capacity)



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

from simulatte.picking_cell.areas import ObservableArea


if TYPE_CHECKING:
    from simulatte.system import System
    from simulatte.picking_cell import PickingCell


T = TypeVar("T", bound=ObservableArea)


class Observer(Generic[T]):
    """
    An observer is in charge of observing a specific logical area of a picking cell.

    It implements a signal event, which is triggered by other objects when there is
    the need to notify the observer about the availability of the observed area.
    """

    def __init__(self, *, system: System, observable_area: T, register_main_process: bool = True) -> None:
        self.system = system
        self.observable_area = observable_area
        if register_main_process:
            self._main = self.system.env.process(self.run())

    @property
    def cell(self) -> PickingCell:
        return self.observable_area.cell

    def run(self):
        """
        Run the observer main process.
        Each observer waits for the signal event to be triggered from the assigned observable area.
        Once triggered, the observer will execute its main process.
        """
        while True:
            yield self.observable_area.signal_event
            self._main_process()

    def next(self):
        """
        Used to get the next item from the observable area.
        """
        raise NotImplementedError

    def _main_process(self):
        """
        The main process of the observer.
        """
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING

from simulatte.exceptions import OutOfStockError
from simulatte.products import Product
from simulatte.stores import WarehouseStore
from simulatte.stores.warehouse_location import PhysicalPosition

if TYPE_CHECKING:
    from simulatte.stores.warehouse_location.warehouse_location import WarehouseLocation


class UnitLoadPolicy:
    def __call__(
        self,
        *,
        stores: list[WarehouseStore],
        product: Product,
        quantity: int,
    ) -> tuple[tuple[WarehouseStore, WarehouseLocation, PhysicalPosition],...]:
        raise NotImplementedError


class ClosestUnitLoadPolicy(UnitLoadPolicy):
    def __call__(
        self,
        *,
        store: WarehouseStore,
        product: Product,
        quantity: int,  # [cases]
    ) -> tuple[WarehouseLocation, PhysicalPosition] | tuple[None, None]:
        locs = sorted(
            (
                location
                for location in store.locations
                if not location.is_empty and not location.fully_booked and location.product == product
            ),
            key=lambda l: (l.n_unit_loads, l.first_available_unit_load.n_cases),
        )

        for location in locs:
            first_position = location.first_position
            unit_load = first_position.unit_load
            if unit_load is not None:  # c'è qualcosa in prima locazione
                if unit_load not in location.booked_pickups:
                    if first_position.n_cases >= quantity:
                        return location, first_position
                else:
                    # se la unit_load nella prima posizione è stata prenotata
                    # non si può accedere alla seconda posizione
                    # quindi la locazione non viene più considerata
                    continue
            else:
                second_position = location.second_position
                unit_load = second_position.unit_load
                if unit_load is not None and unit_load not in location.booked_pickups:
                    if second_position.n_cases >= quantity:
                        return location, second_position

        return None, None


class MultiStoreLocationPolicy(UnitLoadPolicy):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.counter = {
            "best_case": 0,
            "first_aggregation": 0,
            "second_aggregation": 0,
            "magic": 0,
        }

    def __call__(
        self,
        *,
        stores: list[WarehouseStore],
        product: Product,
        quantity: int,  # [cases]
    ) -> tuple[tuple[WarehouseStore, WarehouseLocation, PhysicalPosition],...]:

        # individuiamo le location di tutti gli store che contengono il prodotto richiesto
        # e che non sono completamente prenotate
        locations = [
            location
            for store in stores
            for location in store.locations
            if not location.is_empty and not location.fully_booked and location.product == product
        ]

        # separiamo le unit load che sono in posizione seconda e quelle che sono in posizione prima
        unit_loads_from_half_empty_locations = [
            location.second_position.unit_load
            for location in locations
            if location.is_half_full
            and location.second_position.unit_load is not None
            and location.second_position.unit_load not in location.booked_pickups
        ]
        unit_loads_from_full_locations = [
            location.first_position.unit_load
            if location.first_position.unit_load is not None
            and location.first_position.unit_load not in location.booked_pickups
            else location.second_position.unit_load
            for location in locations
            if location.is_full
        ]

        unit_loads = sorted(
            unit_loads_from_half_empty_locations + unit_loads_from_full_locations,
            key=lambda u: u.n_cases,
        )

        def returner(unit_load) -> tuple[WarehouseStore, WarehouseLocation, PhysicalPosition]:
            position = (
                unit_load.location.first_position
                if unit_load == unit_load.location.first_position.unit_load
                else unit_load.location.second_position
            )
            if unit_load in unit_load.location.booked_pickups:
                raise ValueError("unit load is booked for pickup")
            return (unit_load.location.store, unit_load.location, position)

        # best case: ritorniamo la prima unit load che soddisfa esattamente la quantità richiesta
        for unit_load in unit_loads:
            if unit_load.upper_layer.n_cases == quantity:
                self.counter["best_case"] += 1
                return (returner(unit_load),)

        first_unit_load = unit_loads[0]
        if first_unit_load.n_cases >= quantity:
            return (returner(first_unit_load),)

        second_unit_load = None
        for other_unit_load in unit_loads[1:]:
            if first_unit_load.n_cases + other_unit_load.n_cases == quantity:
                second_unit_load = other_unit_load
                self.counter["first_aggregation"] += 1
                break
        else:
            last_unit_load = unit_loads[-1]
            if first_unit_load.n_cases + last_unit_load.n_cases > quantity:
                self.counter["second_aggregation"] += 1
                second_unit_load = last_unit_load

        if second_unit_load is not None:
            return returner(first_unit_load), returner(second_unit_load)

        raise OutOfStockError(f"no unit load found for {quantity} cases of {product}")



from __future__ import annotations

from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:
    from simulatte import System
    from simulatte.distance import Distance
    from simulatte.location import Location


class DistanceManager:
    def __init__(self, *, system: System, DistanceClass: Type[Distance]) -> None:
        self.system = system
        self.DistanceClass = DistanceClass

    def __call__(self, from_: Location, to: Location) -> Distance:
        return self.DistanceClass(self.system, from_, to)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from simulatte.products import Product


class Stock:
    def __init__(self, *, product: Product, n_cases: int) -> None:
        self.product = product
        self.n_cases = n_cases

    @property
    def n_layers(self) -> float:
        return self.n_cases / self.product.cases_per_layer

    @property
    def n_pallets(self) -> float:
        return self.n_layers / self.product.layers_per_pallet


class OnHand(Stock):
    pass


class OnOrder(Stock):
    pass



from __future__ import annotations

from simpy.core import BoundClass

from .filter_multi_store_get import FilterMultiStoreGet
from ..multi_store import MultiStore


class FilterMultiStore(MultiStore):
    """
    The FilterMultiStore is an extension to the MultiStore which allows the
    storage and retrieval of multiple items at once based on a 'filter' callable.
    """

    get = BoundClass(FilterMultiStoreGet)

    def _do_get(self, event: FilterMultiStoreGet) -> None:
        to_retrieve = []
        for item in self.items:
            if event.filter(item):
                self.items.remove(item)
                to_retrieve.append(item)

        if len(to_retrieve) > 0:
            event.succeed(to_retrieve)
